# Copyright (c) 2022 NVIDIA
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Used by get_swift_info and register_swift_info to store information about
# the swift cluster.
import os
import shutil
import time
import unittest
from collections import defaultdict
from tempfile import mkdtemp

from unittest import mock

from swift.common.middleware import backend_ratelimit
from swift.common.middleware.backend_ratelimit import \
    BackendRateLimitMiddleware
from swift.common.swob import Request, HTTPOk
from test.debug_logger import debug_logger
from test.unit.common.middleware.helpers import FakeSwift


class FakeApp(object):
    def __init__(self):
        self.calls = []

    def __call__(self, env, start_response):
        start_response('200 OK', {})
        return ['']


class TestBackendRatelimitMiddleware(unittest.TestCase):
    def setUp(self):
        super(TestBackendRatelimitMiddleware, self).setUp()
        self.swift = FakeSwift()
        self.tempdir = mkdtemp()
        self.default_req_per_dev_per_sec = dict(
            (key, 0.0) for key in
            (None, 'GET', 'HEAD', 'PUT', 'POST', 'DELETE', 'UPDATE',
             'REPLICATE')
        )

    def tearDown(self):
        shutil.rmtree(self.tempdir, ignore_errors=True)

    def test_init(self):
        conf = {'swift_dir': self.tempdir}
        factory = backend_ratelimit.filter_factory(conf)
        rl = factory(self.swift)
        self.assertEqual(self.default_req_per_dev_per_sec,
                         rl.requests_per_device_per_second)
        self.assertEqual(1.0, rl.requests_per_device_rate_buffer)
        self.assertFalse(rl.is_any_rate_limit_configured)

        conf = {'swift_dir': self.tempdir,
                'requests_per_device_per_second': 1.3,
                'requests_per_device_rate_buffer': 2.4}
        factory = backend_ratelimit.filter_factory(conf)
        rl = factory(self.swift)
        exp_req_per_dev_per_sec = dict(self.default_req_per_dev_per_sec)
        exp_req_per_dev_per_sec[None] = 1.3
        self.assertEqual(exp_req_per_dev_per_sec,
                         rl.requests_per_device_per_second)
        self.assertEqual(2.4, rl.requests_per_device_rate_buffer)
        self.assertTrue(rl.is_any_rate_limit_configured)

        conf = {'requests_per_device_per_second': -1}
        factory = backend_ratelimit.filter_factory(conf)
        with self.assertRaises(ValueError) as cm:
            factory(self.swift)
        self.assertEqual(
            'Value must be a non-negative float number, not "-1.0".',
            str(cm.exception))

        conf = {'requests_per_device_rate_buffer': -1}
        factory = backend_ratelimit.filter_factory(conf)
        with self.assertRaises(ValueError):
            factory(self.swift)
        self.assertEqual(
            'Value must be a non-negative float number, not "-1.0".',
            str(cm.exception))

    def test_init_conf_path(self):
        conf = {}
        factory = backend_ratelimit.filter_factory(conf)
        rl = factory(self.swift)
        self.assertEqual('/etc/swift/backend-ratelimit.conf', rl.conf_path)
        conf = {'backend_ratelimit_conf_path': '/etc/other/rl.conf'}
        factory = backend_ratelimit.filter_factory(conf)
        rl = factory(self.swift)
        self.assertEqual('/etc/other/rl.conf', rl.conf_path)
        conf = {'backend_ratelimit_conf_path': ''}
        factory = backend_ratelimit.filter_factory(conf)
        rl = factory(self.swift)
        self.assertEqual('', rl.conf_path)

    def test_init_conf_reload_interval(self):
        conf = {}
        factory = backend_ratelimit.filter_factory(conf)
        rl = factory(self.swift)
        self.assertEqual(60, rl.config_reload_interval)
        conf = {'config_reload_interval': 600}
        factory = backend_ratelimit.filter_factory(conf)
        rl = factory(self.swift)
        self.assertEqual(600, rl.config_reload_interval)
        conf = {'config_reload_interval': 0}
        factory = backend_ratelimit.filter_factory(conf)
        rl = factory(self.swift)
        self.assertEqual(0, rl.config_reload_interval)

        def test_bad(value):
            with self.assertRaises(ValueError) as cm:
                conf = {'config_reload_interval': value}
                factory = backend_ratelimit.filter_factory(conf)
                factory(self.swift)
            self.assertIn('Value must be a non-negative float number',
                          str(cm.exception))
        test_bad(-1)
        test_bad('auto')

    def test_init_config_file_set_and_missing(self):
        # warn if missing conf file during init (conf_path set)
        def do_test(conf_path):
            conf = {'backend_ratelimit_conf_path': '%s' % conf_path,
                    'requests_per_device_per_second': "1.3"}
            factory = backend_ratelimit.filter_factory(conf)
            with mock.patch(
                    'swift.common.middleware.backend_ratelimit.get_logger',
                    return_value=debug_logger()):
                rl = factory(self.swift)
            exp_req_per_dev_per_sec = dict(self.default_req_per_dev_per_sec)
            exp_req_per_dev_per_sec.update({None: 1.3})
            self.assertEqual(exp_req_per_dev_per_sec,
                             rl.requests_per_device_per_second)
            self.assertEqual(1.0, rl.requests_per_device_rate_buffer)
            self.assertEqual([], rl.logger.get_lines_for_level('error'))
            self.assertEqual(
                ['Failed to load config file, config unchanged: Unable to '
                 'read config from %s' % conf_path],
                rl.logger.get_lines_for_level('warning'))

        do_test('')
        do_test(os.path.join(self.tempdir, 'backend_rl.conf'))

    def test_init_config_file_unset_and_missing(self):
        # don't warn if missing conf file during init (conf_path not set)
        conf = {'swift_dir': self.tempdir,
                'requests_per_device_per_second': "1.3"}
        factory = backend_ratelimit.filter_factory(conf)
        with mock.patch(
                'swift.common.middleware.backend_ratelimit.get_logger',
                return_value=debug_logger()):
            rl = factory(self.swift)
        exp_req_per_dev_per_sec = dict(self.default_req_per_dev_per_sec)
        exp_req_per_dev_per_sec[None] = 1.3
        self.assertEqual(exp_req_per_dev_per_sec,
                         rl.requests_per_device_per_second)
        self.assertEqual(1.0, rl.requests_per_device_rate_buffer)
        self.assertEqual([], rl.logger.get_lines_for_level('error'))
        self.assertEqual([], rl.logger.get_lines_for_level('warning'))

    def test_init_config_file_no_section(self):
        # warn and ignore conf file without section
        conf_path = os.path.join(self.tempdir, 'backend_rl.conf')
        with open(conf_path, 'w') as fd:
            fd.write('[DEFAULT]\n'
                     'requests_per_device_per_second = 12.3\n')
        conf = {'backend_ratelimit_conf_path': '%s' % conf_path,
                'requests_per_device_per_second': 1.3}
        factory = backend_ratelimit.filter_factory(conf)
        with mock.patch('swift.common.middleware.backend_ratelimit.get_logger',
                        return_value=debug_logger()):
            rl = factory(self.swift)
        exp_req_per_dev_per_sec = dict(self.default_req_per_dev_per_sec)
        exp_req_per_dev_per_sec[None] = 1.3
        self.assertEqual(exp_req_per_dev_per_sec,
                         rl.requests_per_device_per_second)
        self.assertEqual(1.0, rl.requests_per_device_rate_buffer)
        lines = rl.logger.get_lines_for_level('warning')
        self.assertEqual(1, len(lines), lines)
        self.assertIn('Invalid config file', lines[0])
        self.assertEqual([], rl.logger.get_lines_for_level('error'))

    def test_read_default_backend_ratelimit_conf(self):
        conf = {'swift_dir': self.tempdir,
                'requests_per_device_per_second': "1.3",
                'requests_per_device_rate_buffer': "2.4",
                # do not set 'backend_ratelimit_conf_path'
                'config_reload_interval': 15}
        # but set it up anyway
        conf_path = os.path.join(self.tempdir, 'backend-ratelimit.conf')
        with open(conf_path, 'w') as fd:
            fd.write('[backend_ratelimit]\n'
                     'requests_per_device_per_second = 12.3\n')
        factory = backend_ratelimit.filter_factory(conf)
        with mock.patch('swift.common.middleware.backend_ratelimit.get_logger',
                        return_value=debug_logger()):
            rl = factory(self.swift)
        # backend-ratelimit.conf overrides options
        exp_req_per_dev_per_sec = dict(self.default_req_per_dev_per_sec)
        exp_req_per_dev_per_sec[None] = 12.3
        self.assertEqual(exp_req_per_dev_per_sec,
                         rl.requests_per_device_per_second)
        # but only the ones that are listed
        self.assertEqual(2.4, rl.requests_per_device_rate_buffer)
        self.assertTrue(rl.is_any_rate_limit_configured)
        lines = rl.logger.get_lines_for_level('info')
        self.assertEqual(['Loaded config file %s, config changed' % conf_path],
                         lines)

    def test_config_reload_does_not_override_reload_options(self):
        conf_path = os.path.join(self.tempdir, 'override-ratelimit.conf')
        conf = {'swift_dir': self.tempdir,
                'requests_per_device_per_second': "1.3",
                'requests_per_device_rate_buffer': "2.4",
                'backend_ratelimit_conf_path': conf_path,
                'config_reload_interval': 15}
        with open(conf_path, 'w') as fd:
            fd.write('[backend_ratelimit]\n'
                     'requests_per_device_per_second = 12.3\n'
                     'requests_per_device_rate_buffer = 12.4\n'
                     'backend_ratelimit_conf_path = /etc/swift/ignored.conf\n'
                     'config_reload_interval = 999999\n')
        factory = backend_ratelimit.filter_factory(conf)
        with mock.patch('swift.common.middleware.backend_ratelimit.get_logger',
                        return_value=debug_logger()):
            rl = factory(self.swift)
        exp_req_per_dev_per_sec = dict(self.default_req_per_dev_per_sec)
        exp_req_per_dev_per_sec[None] = 12.3
        self.assertEqual(exp_req_per_dev_per_sec,
                         rl.requests_per_device_per_second)
        self.assertEqual(12.4, rl.requests_per_device_rate_buffer)
        self.assertTrue(rl.is_any_rate_limit_configured)
        # options related to conf file loading are not loaded from conf file...
        self.assertEqual(conf_path, rl.conf_path)
        self.assertEqual(15, rl.config_reload_interval)
        lines = rl.logger.logger.get_lines_for_level('info')
        self.assertEqual(['Loaded config file %s, config changed' % conf_path],
                         lines)

    def _do_test_init_config_file_overrides_filter_conf(
            self, path_to_actual_conf_file, configured_conf_path):
        # verify that conf file options override filter conf options
        # create the actual file, but no options
        with open(path_to_actual_conf_file, 'w') as fd:
            fd.write('[backend_ratelimit]')
        conf = {'swift_dir': self.tempdir,
                'requests_per_device_per_second': "1.3",
                'requests_per_device_rate_buffer': "2.4",
                'config_reload_interval': 15}
        if configured_conf_path:
            # only configure if given a conf_path
            conf['backend_ratelimit_conf_path'] = configured_conf_path
            exp_configured_conf_path = configured_conf_path
        else:
            # fall back to default
            exp_configured_conf_path = os.path.join(self.tempdir,
                                                    'backend-ratelimit.conf')
        factory = backend_ratelimit.filter_factory(conf)
        rl = factory(self.swift)
        exp_req_per_dev_per_sec = dict(self.default_req_per_dev_per_sec)
        exp_req_per_dev_per_sec[None] = 1.3
        self.assertEqual(exp_req_per_dev_per_sec,
                         rl.requests_per_device_per_second)
        self.assertEqual(2.4, rl.requests_per_device_rate_buffer)
        self.assertTrue(rl.is_any_rate_limit_configured)
        self.assertEqual(exp_configured_conf_path, rl.conf_path)
        self.assertEqual(15, rl.config_reload_interval)

        # create file with option
        with open(path_to_actual_conf_file, 'w') as fd:
            fd.write('[backend_ratelimit]\n'
                     'requests_per_device_per_second = 12.3\n'
                     'backend_ratelimit_conf_path = /etc/swift/ignored.conf\n'
                     'config_reload_interval = 999999\n')
        factory = backend_ratelimit.filter_factory(conf)
        rl = factory(self.swift)
        exp_req_per_dev_per_sec[None] = 12.3
        self.assertEqual(exp_req_per_dev_per_sec,
                         rl.requests_per_device_per_second)
        self.assertEqual(2.4, rl.requests_per_device_rate_buffer)
        self.assertTrue(rl.is_any_rate_limit_configured)
        # options related to conf file loading are not loaded from conf file...
        self.assertEqual(exp_configured_conf_path, rl.conf_path)
        self.assertEqual(15, rl.config_reload_interval)

        with open(path_to_actual_conf_file, 'w') as fd:
            fd.write(
                '[backend_ratelimit]\n'
                'requests_per_device_per_second = 5.3\n'
                'requests_per_device_rate_buffer = 0.5\n'
                'delete_requests_per_device_per_second = 1\n'
                'get_requests_per_device_per_second = 2\n'
                'head_requests_per_device_per_second = 3\n'
                'post_requests_per_device_per_second = 4\n'
                'put_requests_per_device_per_second = 5\n'
                'replicate_requests_per_device_per_second = 6\n'
                'update_requests_per_device_per_second = 7\n'
                'backend_ratelimit_conf_path = /etc/swift/ignored.conf\n'
                'config_reload_interval = 999999\n'
            )
        factory = backend_ratelimit.filter_factory(conf)
        rl = factory(self.swift)
        exp_req_per_dev_per_sec.update(
            {
                None: 5.3,
                'DELETE': 1,
                'GET': 2,
                'HEAD': 3,
                'POST': 4,
                'PUT': 5,
                'REPLICATE': 6,
                'UPDATE': 7,
            }
        )
        self.assertEqual(exp_req_per_dev_per_sec,
                         rl.requests_per_device_per_second)
        self.assertEqual(0.5, rl.requests_per_device_rate_buffer)
        self.assertTrue(rl.is_any_rate_limit_configured)
        # options related to conf file loading are not loaded from conf file...
        self.assertEqual(exp_configured_conf_path, rl.conf_path)
        self.assertEqual(15, rl.config_reload_interval)

    def test_init_config_file_at_default_path_overrides_filter_conf(self):
        # default conf path is loaded if it exists
        default_conf_path = os.path.join(self.tempdir,
                                         'backend-ratelimit.conf')
        self._do_test_init_config_file_overrides_filter_conf(
            path_to_actual_conf_file=default_conf_path,
            configured_conf_path=None)

        self._do_test_init_config_file_overrides_filter_conf(
            path_to_actual_conf_file=default_conf_path,
            configured_conf_path=default_conf_path)

    def test_init_config_file_at_configured_path_overrides_filter_conf(self):
        # explicitly configured conf path is loaded
        custom_conf_path = os.path.join(self.tempdir, 'backend_rl.conf')
        self._do_test_init_config_file_overrides_filter_conf(
            path_to_actual_conf_file=custom_conf_path,
            configured_conf_path=custom_conf_path)

    def _do_test_config_file_reload(self, reload_interval):
        # verify that conf file options are periodically reloaded
        filter_conf = {'swift_dir': self.tempdir,
                       'requests_per_device_per_second': "1.3",
                       'requests_per_device_rate_buffer': "2.4",
                       'head_requests_per_device_per_second': '6.2'}
        if reload_interval:
            filter_conf['config_reload_interval'] = reload_interval

        now = time.time()
        # create the actual file
        conf_path = os.path.join(filter_conf['swift_dir'],
                                 'backend-ratelimit.conf')
        with open(conf_path, 'w') as fd:
            fd.write('[backend_ratelimit]\n'
                     'requests_per_device_per_second = 12.3\n'
                     # conf file cannot re-configure where the conf file is...
                     'backend_ratelimit_conf_path = /etc/ignored\n'
                     'config_reload_interval = also_ignored\n')
        factory = backend_ratelimit.filter_factory(filter_conf)
        with mock.patch('swift.common.middleware.backend_ratelimit.time.time',
                        return_value=now):
            rl = factory(self.swift)
        exp_req_per_dev_per_sec = dict(self.default_req_per_dev_per_sec)
        exp_req_per_dev_per_sec.update({None: 12.3, 'HEAD': float(
            filter_conf['head_requests_per_device_per_second'])})
        self.assertEqual(exp_req_per_dev_per_sec,
                         rl.requests_per_device_per_second)
        self.assertEqual(float(filter_conf['requests_per_device_rate_buffer']),
                         rl.requests_per_device_rate_buffer)
        self.assertEqual(conf_path, rl.conf_path)

        # modify the conf file
        with open(conf_path, 'w') as fd:
            fd.write('[backend_ratelimit]\n'
                     'requests_per_device_per_second = 29.3\n'
                     'requests_per_device_rate_buffer = 12.4\n'
                     'backend_ratelimit_conf_path = /etc/ignored\n'
                     'config_reload_interval = also_ignored\n'
                     'head_requests_per_device_per_second = 5.1\n'
                     'delete_requests_per_device_per_second = 7.3\n'
                     'get_requests_per_device_per_second = 8.4\n')

        # send some requests, but too soon for config file to be reloaded
        req1 = Request.blank('/sda1/99/a/c/o')
        req2 = Request.blank('/sda2/99/a/c/o',
                             environ={'REQUEST_METHOD': 'DELETE'})
        self.swift.register(req1.method, req1.path, HTTPOk, {})
        self.swift.register(req2.method, req2.path, HTTPOk, {})
        with mock.patch('swift.common.middleware.backend_ratelimit.time.time',
                        return_value=now + rl.config_reload_interval - 1):
            resp1 = req1.get_response(rl)
            resp2 = req2.get_response(rl)
        self.assertEqual(200, resp1.status_int)
        self.assertEqual(200, resp2.status_int)
        self.assertEqual(exp_req_per_dev_per_sec,
                         rl.requests_per_device_per_second)
        self.assertEqual(float(filter_conf['requests_per_device_rate_buffer']),
                         rl.requests_per_device_rate_buffer)
        self.assertEqual(conf_path, rl.conf_path)

        # verify the per dev ratelimiters
        self.assertEqual({('sda1', 'GET'): 0.0,
                          ('sda2', 'DELETE'): 0.0,
                          ('sda1', None): 12.3,
                          ('sda2', None): 12.3},
                         dict((key, val.max_rate)
                              for key, val in rl.rate_limiters.items()))
        for (dev, method), limiter in rl.rate_limiters.items():
            self.assertEqual(2.4 * limiter.clock_accuracy,
                             limiter.rate_buffer_ms, (dev, method))

        # send some requests, time for config file to be reloaded
        with mock.patch('swift.common.middleware.backend_ratelimit.time.time',
                        return_value=now + rl.config_reload_interval + 0.01):
            resp1 = req1.get_response(rl)
            resp2 = req2.get_response(rl)
        self.assertEqual(200, resp1.status_int)
        self.assertEqual(200, resp2.status_int)
        exp_req_per_dev_per_sec.update({
            None: 29.3,
            'HEAD': 5.1,
            'DELETE': 7.3,
            'GET': 8.4})
        self.assertEqual(exp_req_per_dev_per_sec,
                         rl.requests_per_device_per_second)
        self.assertEqual(12.4, rl.requests_per_device_rate_buffer)
        self.assertEqual(conf_path, rl.conf_path)

        # verify the per dev ratelimiters were updated
        self.assertEqual({('sda1', 'GET'): 8.4,
                          ('sda2', 'DELETE'): 7.3,
                          ('sda1', None): 29.3,
                          ('sda2', None): 29.3},
                         dict((key, val.max_rate)
                              for key, val in rl.rate_limiters.items()))
        for (dev, method), limiter in rl.rate_limiters.items():
            self.assertEqual(12.4 * limiter.clock_accuracy,
                             limiter.rate_buffer_ms, (dev, method))

        # modify the config file again
        # remove requests_per_device_per_second option
        # remove [head|delete]_requests_per_device_per_second options
        with open(conf_path, 'w') as fd:
            fd.write('[backend_ratelimit]\n'
                     'backend_ratelimit_conf_path = /etc/ignored\n'
                     'config_reload_interval = also_ignored\n'
                     'requests_per_device_rate_buffer = 0.5\n'
                     'get_requests_per_device_per_second = 9.5\n')

        # send some requests, not yet time for config file to be reloaded
        with mock.patch('swift.common.middleware.backend_ratelimit.time.time',
                        return_value=now + 2 * rl.config_reload_interval - 1):
            resp1 = req1.get_response(rl)
            resp2 = req2.get_response(rl)
        self.assertEqual(200, resp1.status_int)
        self.assertEqual(200, resp2.status_int)
        self.assertEqual(exp_req_per_dev_per_sec,
                         rl.requests_per_device_per_second)
        self.assertEqual(12.4, rl.requests_per_device_rate_buffer)
        self.assertEqual(conf_path, rl.conf_path)

        # verify the per dev ratelimiters were not updated
        self.assertEqual({('sda1', 'GET'): 8.4,
                          ('sda2', 'DELETE'): 7.3,
                          ('sda1', None): 29.3,
                          ('sda2', None): 29.3},
                         dict((key, val.max_rate)
                              for key, val in rl.rate_limiters.items()))
        for (dev, method), limiter in rl.rate_limiters.items():
            self.assertEqual(12.4 * limiter.clock_accuracy,
                             limiter.rate_buffer_ms, (dev, method))

        # send some requests, time for config file to be reloaded
        with mock.patch(
                'swift.common.middleware.backend_ratelimit.time.time',
                return_value=now + 2 * rl.config_reload_interval + 0.01):
            resp1 = req1.get_response(rl)
            resp2 = req2.get_response(rl)
        self.assertEqual(200, resp1.status_int)
        self.assertEqual(200, resp2.status_int)
        # requests_per_device_per_second option reverts to filter conf
        # delete_requests_per_device_per_second option reverts to default
        # head_requests_per_device_per_second option reverts to filter conf
        exp_req_per_dev_per_sec.update({
            None: 1.3,
            'HEAD': 6.2,
            'DELETE': 0.0,
            'GET': 9.5})
        self.assertEqual(exp_req_per_dev_per_sec,
                         rl.requests_per_device_per_second)
        self.assertEqual(0.5, rl.requests_per_device_rate_buffer)
        self.assertEqual(conf_path, rl.conf_path)

        # verify the per dev ratelimiters were updated
        self.assertEqual({('sda1', 'GET'): 9.5,
                          ('sda2', 'DELETE'): 0.0,
                          ('sda1', None): 1.3,
                          ('sda2', None): 1.3},
                         dict((key, val.max_rate)
                              for key, val in rl.rate_limiters.items()))
        for (dev, method), limiter in rl.rate_limiters.items():
            self.assertEqual(0.5 * limiter.clock_accuracy,
                             limiter.rate_buffer_ms, (dev, method))
        return rl

    def test_config_file_reload_default_interval(self):
        rl = self._do_test_config_file_reload(None)
        self.assertEqual(60, rl.config_reload_interval)

    def test_config_file_reload_custom_interval(self):
        rl = self._do_test_config_file_reload(30.1)
        self.assertEqual(30.1, rl.config_reload_interval)

    def test_config_file_reload_clears_all_limits(self):
        # verify that reloaded config file can disable all rate limits
        now = time.time()
        conf_path = os.path.join(self.tempdir, 'missing')
        filter_conf = {'swift_dir': self.tempdir,
                       # path set so expect warning during init
                       'backend_ratelimit_conf_path': conf_path,
                       'requests_per_device_per_second': "1.3",
                       'head_requests_per_device_per_second = 1.1\n'
                       'requests_per_device_rate_buffer': "2.4"}
        with open(conf_path, 'w') as fd:
            fd.write('[backend_ratelimit]\n'
                     'requests_per_device_per_second = 29.3\n'
                     'head_requests_per_device_per_second = 5.1\n'
                     'get_requests_per_device_per_second = 8.4\n')
        factory = backend_ratelimit.filter_factory(filter_conf)

        # expect warning during init
        with mock.patch('swift.common.middleware.backend_ratelimit.time.time',
                        return_value=now):
            with mock.patch(
                    'swift.common.middleware.backend_ratelimit.get_logger',
                    return_value=debug_logger()):
                rl = factory(self.swift)
        # filter conf has been applied
        exp_req_per_dev_per_sec = dict(self.default_req_per_dev_per_sec)
        exp_req_per_dev_per_sec.update({
            None: 29.3,
            'HEAD': 5.1,
            'GET': 8.4})
        self.assertTrue(rl.is_any_rate_limit_configured)
        self.assertEqual(exp_req_per_dev_per_sec,
                         rl.requests_per_device_per_second)
        self.assertEqual([], rl.logger.get_lines_for_level('warning'))
        self.assertEqual([], rl.logger.get_lines_for_level('error'))

        # write zero rate limits to conf file
        # jump into future, send request, config reload attempted
        with open(conf_path, 'w') as fd:
            fd.write('[backend_ratelimit]\n'
                     'requests_per_device_per_second = 0.0\n'
                     'head_requests_per_device_per_second = 0.0\n'
                     'get_requests_per_device_per_second = 0.0\n')
        req1 = Request.blank('/sda1/99/a/c/o')
        self.swift.register(req1.method, req1.path, HTTPOk, {})
        with mock.patch('swift.common.middleware.backend_ratelimit.time.time',
                        return_value=now + 10000):
            resp1 = req1.get_response(rl)
        self.assertEqual(200, resp1.status_int)
        exp_req_per_dev_per_sec = dict(self.default_req_per_dev_per_sec)
        self.assertFalse(rl.is_any_rate_limit_configured)
        self.assertEqual(exp_req_per_dev_per_sec,
                         rl.requests_per_device_per_second)
        self.assertEqual([], rl.logger.get_lines_for_level('warning'))
        self.assertEqual([], rl.logger.get_lines_for_level('error'))

    def test_config_file_reload_set_and_missing(self):
        now = time.time()
        conf_path = os.path.join(self.tempdir, 'missing')
        filter_conf = {'swift_dir': self.tempdir,
                       # path set so expect warning during init
                       'backend_ratelimit_conf_path': conf_path,
                       'requests_per_device_per_second': "1.3",
                       'requests_per_device_rate_buffer': "2.4"}
        factory = backend_ratelimit.filter_factory(filter_conf)

        # expect warning during init
        with mock.patch('swift.common.middleware.backend_ratelimit.time.time',
                        return_value=now):
            with mock.patch(
                    'swift.common.middleware.backend_ratelimit.get_logger',
                    return_value=debug_logger()):
                rl = factory(self.swift)
        # filter conf has been applied
        exp_req_per_dev_per_sec = dict(self.default_req_per_dev_per_sec)
        exp_req_per_dev_per_sec[None] = 1.3
        self.assertTrue(rl.is_any_rate_limit_configured)
        self.assertEqual(exp_req_per_dev_per_sec,
                         rl.requests_per_device_per_second)
        self.assertEqual(2.4, rl.requests_per_device_rate_buffer)
        self.assertEqual(
            ['Failed to load config file, config unchanged: Unable to read '
             'config from %s' % conf_path],
            rl.logger.get_lines_for_level('warning'))
        self.assertEqual([], rl.logger.get_lines_for_level('error'))

        # jump into future, send request, config reload attempted
        # no ongoing warning
        rl.logger.logger.clear()
        req1 = Request.blank('/sda1/99/a/c/o')
        self.swift.register(req1.method, req1.path, HTTPOk, {})
        with mock.patch('swift.common.middleware.backend_ratelimit.time.time',
                        return_value=now + 10000):
            resp1 = req1.get_response(rl)
        self.assertEqual(200, resp1.status_int)
        self.assertTrue(rl.is_any_rate_limit_configured)
        self.assertEqual(exp_req_per_dev_per_sec,
                         rl.requests_per_device_per_second)
        self.assertEqual(2.4, rl.requests_per_device_rate_buffer)
        self.assertEqual([], rl.logger.get_lines_for_level('warning'))
        self.assertEqual([], rl.logger.get_lines_for_level('error'))

    def test_config_file_reload_unset_and_missing(self):
        now = time.time()
        filter_conf = {'swift_dir': self.tempdir,
                       # conf path not set so expect no warnings
                       'requests_per_device_per_second': "1.3",
                       'requests_per_device_rate_buffer': "2.4"}
        factory = backend_ratelimit.filter_factory(filter_conf)

        # expect NO warning during init
        with mock.patch('swift.common.middleware.backend_ratelimit.time.time',
                        return_value=now):
            with mock.patch(
                    'swift.common.middleware.backend_ratelimit.get_logger',
                    return_value=debug_logger()):
                rl = factory(self.swift)
        # filter conf has been applied
        exp_req_per_dev_per_sec = dict(self.default_req_per_dev_per_sec)
        exp_req_per_dev_per_sec[None] = 1.3
        self.assertEqual(exp_req_per_dev_per_sec,
                         rl.requests_per_device_per_second)
        self.assertEqual(2.4, rl.requests_per_device_rate_buffer)
        self.assertEqual([], rl.logger.get_lines_for_level('warning'))
        self.assertEqual([], rl.logger.get_lines_for_level('error'))

        # jump into future, send request, config reload attempted
        # no ongoing warning
        req1 = Request.blank('/sda1/99/a/c/o')
        self.swift.register(req1.method, req1.path, HTTPOk, {})
        with mock.patch('swift.common.middleware.backend_ratelimit.time.time',
                        return_value=now + 10000):
            resp1 = req1.get_response(rl)
        self.assertEqual(200, resp1.status_int)
        # previous conf file value has been retained
        self.assertEqual(exp_req_per_dev_per_sec,
                         rl.requests_per_device_per_second)
        self.assertEqual(2.4, rl.requests_per_device_rate_buffer)
        self.assertEqual([], rl.logger.get_lines_for_level('warning'))
        self.assertEqual([], rl.logger.get_lines_for_level('error'))

    def test_config_file_reload_empty_section(self):
        # verify that empty section is OK
        now = time.time()
        filter_conf = {'swift_dir': self.tempdir,
                       'requests_per_device_per_second': "1.3",
                       'requests_per_device_rate_buffer': "2.4"}
        # create the actual file
        conf_path = os.path.join(self.tempdir, 'backend-ratelimit.conf')
        with open(conf_path, 'w') as fd:
            fd.write('[backend_ratelimit]\n')
        factory = backend_ratelimit.filter_factory(filter_conf)
        with mock.patch('swift.common.middleware.backend_ratelimit.time.time',
                        return_value=now):
            with mock.patch(
                    'swift.common.middleware.backend_ratelimit.get_logger',
                    return_value=debug_logger()):
                rl = factory(self.swift)
        # conf file value has been applied
        exp_req_per_dev_per_sec = dict(self.default_req_per_dev_per_sec)
        exp_req_per_dev_per_sec[None] = 1.3
        self.assertEqual(exp_req_per_dev_per_sec,
                         rl.requests_per_device_per_second)
        self.assertEqual(2.4, rl.requests_per_device_rate_buffer)
        self.assertEqual([], rl.logger.get_lines_for_level('warning'))
        self.assertEqual([], rl.logger.get_lines_for_level('error'))

    def test_config_file_reload_error(self):
        # verify that current config is preserved if reload fails
        now = time.time()
        filter_conf = {'swift_dir': self.tempdir,
                       'requests_per_device_per_second': "1.3",
                       'requests_per_device_rate_buffer': "2.4"}
        # create the actual file
        conf_path = os.path.join(self.tempdir, 'backend-ratelimit.conf')
        with open(conf_path, 'w') as fd:
            fd.write('[backend_ratelimit]\n'
                     'requests_per_device_per_second = 12.3\n')
        factory = backend_ratelimit.filter_factory(filter_conf)
        with mock.patch('swift.common.middleware.backend_ratelimit.time.time',
                        return_value=now):
            with mock.patch(
                    'swift.common.middleware.backend_ratelimit.get_logger',
                    return_value=debug_logger()):
                rl = factory(self.swift)
        # conf file value has been applied
        exp_req_per_dev_per_sec = dict(self.default_req_per_dev_per_sec)
        exp_req_per_dev_per_sec[None] = 12.3
        self.assertEqual(exp_req_per_dev_per_sec,
                         rl.requests_per_device_per_second)
        self.assertEqual(2.4, rl.requests_per_device_rate_buffer)
        self.assertEqual([], rl.logger.get_lines_for_level('warning'))
        self.assertEqual([], rl.logger.get_lines_for_level('error'))

        with open(conf_path, 'w') as fd:
            fd.write('[backend_ratelimit]\n'
                     'requests_per_device_per_second = 29.3\n')

        # jump into future, send request, config reload attempted but fails
        req1 = Request.blank('/sda1/99/a/c/o')
        self.swift.register(req1.method, req1.path, HTTPOk, {})
        with mock.patch('swift.common.middleware.backend_ratelimit.time.time',
                        return_value=now + 10000):
            with mock.patch(
                    'swift.common.middleware.backend_ratelimit.readconf',
                    side_effect=ValueError('BOOM')
            ) as mock_readconf:
                resp1 = req1.get_response(rl)
        self.assertEqual(200, resp1.status_int)
        # previous conf file value has been retained
        self.assertEqual(exp_req_per_dev_per_sec,
                         rl.requests_per_device_per_second)
        self.assertEqual(2.4, rl.requests_per_device_rate_buffer)
        mock_readconf.assert_called_once()
        self.assertEqual(
            ['Invalid config file %s, config unchanged: BOOM' % conf_path],
            rl.logger.get_lines_for_level('warning'))
        self.assertEqual([], rl.logger.get_lines_for_level('error'))

        # the reload is not tried again immediately
        rl.logger = debug_logger()
        with mock.patch('swift.common.middleware.backend_ratelimit.time.time',
                        return_value=now + 10059):
            resp1 = req1.get_response(rl)
        self.assertEqual(200, resp1.status_int)
        # previous conf file value has been retained
        self.assertEqual(exp_req_per_dev_per_sec,
                         rl.requests_per_device_per_second)
        self.assertEqual(2.4, rl.requests_per_device_rate_buffer)
        self.assertEqual([], rl.logger.get_lines_for_level('warning'))
        self.assertEqual([], rl.logger.get_lines_for_level('error'))

        # ..but will be retried after reload interval
        rl.logger = debug_logger()
        with mock.patch('swift.common.middleware.backend_ratelimit.time.time',
                        return_value=now + 10060):
            resp1 = req1.get_response(rl)
        self.assertEqual(200, resp1.status_int)
        # previous conf file value has been retained
        exp_req_per_dev_per_sec.update({None: 29.3})
        self.assertEqual(exp_req_per_dev_per_sec,
                         rl.requests_per_device_per_second)
        self.assertEqual(2.4, rl.requests_per_device_rate_buffer)
        self.assertEqual([], rl.logger.get_lines_for_level('warning'))
        self.assertEqual([], rl.logger.get_lines_for_level('error'))

    def test_config_file_reload_logging(self):
        # verify that config reload is logged when config changes
        now = time.time()
        filter_conf = {'swift_dir': self.tempdir,
                       'requests_per_device_per_second': "1.3",
                       'requests_per_device_rate_buffer': "2.4"}
        # create the actual file
        conf_path = os.path.join(self.tempdir, 'backend-ratelimit.conf')
        with open(conf_path, 'w') as fd:
            fd.write('[backend_ratelimit]\n'
                     'requests_per_device_per_second = 12.3\n')
        factory = backend_ratelimit.filter_factory(filter_conf)
        with mock.patch('swift.common.middleware.backend_ratelimit.time.time',
                        return_value=now):
            with mock.patch(
                    'swift.common.middleware.backend_ratelimit.get_logger',
                    return_value=debug_logger()):
                rl = factory(self.swift)
        # conf file value has been applied
        exp_req_per_dev_per_sec = dict(self.default_req_per_dev_per_sec)
        exp_req_per_dev_per_sec[None] = 12.3
        self.assertEqual(exp_req_per_dev_per_sec,
                         rl.requests_per_device_per_second)
        self.assertEqual(2.4, rl.requests_per_device_rate_buffer)
        lines = rl.logger.get_lines_for_level('info')
        self.assertEqual(['Loaded config file %s, config changed' % conf_path],
                         lines)

        # jump into future, send request, config reload attempted, no change
        rl.logger.logger.clear()
        req1 = Request.blank('/sda1/99/a/c/o')
        self.swift.register(req1.method, req1.path, HTTPOk, {})
        with mock.patch('swift.common.middleware.backend_ratelimit.time.time',
                        return_value=now + 10000):
            resp1 = req1.get_response(rl)
        self.assertEqual(200, resp1.status_int)
        self.assertEqual(exp_req_per_dev_per_sec,
                         rl.requests_per_device_per_second)
        self.assertEqual(2.4, rl.requests_per_device_rate_buffer)
        lines = rl.logger.get_lines_for_level('info')
        self.assertEqual([], lines)

        # modify config file, jump into future, change logged
        with open(conf_path, 'w') as fd:
            fd.write('[backend_ratelimit]\n'
                     'requests_per_device_per_second = 23.4\n')
        rl.logger = debug_logger()
        with mock.patch('swift.common.middleware.backend_ratelimit.time.time',
                        return_value=now + 10060):
            resp1 = req1.get_response(rl)
        self.assertEqual(200, resp1.status_int)
        # previous conf file value has been retained
        exp_req_per_dev_per_sec.update({None: 23.4})
        self.assertEqual(exp_req_per_dev_per_sec,
                         rl.requests_per_device_per_second)
        self.assertEqual(2.4, rl.requests_per_device_rate_buffer)
        lines = rl.logger.get_lines_for_level('info')
        self.assertEqual(['Loaded config file %s, config changed' % conf_path],
                         lines)

    def test_config_file_disappears_appears_logging(self):
        # verify that config reload is logged when file reappears
        now = time.time()
        filter_conf = {'swift_dir': self.tempdir,
                       'requests_per_device_per_second': "1.3",
                       'requests_per_device_rate_buffer': "2.4"}
        # create the actual file
        conf_path = os.path.join(self.tempdir, 'backend-ratelimit.conf')
        conf_str = ('[backend_ratelimit]\n'
                    'requests_per_device_per_second = 12.3\n')
        with open(conf_path, 'w') as fd:
            fd.write(conf_str)
        factory = backend_ratelimit.filter_factory(filter_conf)
        with mock.patch('swift.common.middleware.backend_ratelimit.time.time',
                        return_value=now):
            with mock.patch(
                    'swift.common.middleware.backend_ratelimit.get_logger',
                    return_value=debug_logger()):
                rl = factory(self.swift)
        # conf file value has been applied
        exp_req_per_dev_per_sec = dict(self.default_req_per_dev_per_sec)
        exp_req_per_dev_per_sec[None] = 12.3
        self.assertEqual(exp_req_per_dev_per_sec,
                         rl.requests_per_device_per_second)
        self.assertEqual(2.4, rl.requests_per_device_rate_buffer)
        lines = rl.logger.get_lines_for_level('info')
        self.assertEqual(
            ['Loaded config file %s, config changed' % conf_path],
            lines)
        lines = rl.logger.get_lines_for_level('warning')
        self.assertFalse(lines)

        def do_request(now):
            rl.logger.logger.clear()
            req1 = Request.blank('/sda1/99/a/c/o')
            self.swift.register(req1.method, req1.path, HTTPOk, {})
            with mock.patch(
                    'swift.common.middleware.backend_ratelimit.time.time',
                    return_value=now):
                resp = req1.get_response(rl)
            self.assertEqual(200, resp.status_int)
            info_lines = rl.logger.get_lines_for_level('info')
            warning_lines = rl.logger.get_lines_for_level('warning')
            return info_lines, warning_lines

        # jump into future, send request, config reload fails - warning
        os.unlink(conf_path)
        now += 100
        info_lines, warning_lines = do_request(now)
        self.assertFalse(info_lines)
        self.assertEqual(
            ['Failed to load config file, config unchanged: Unable to '
             'read config from %s' % conf_path], warning_lines)

        # jump into future, send request, config reload fails - no warning
        now += 100
        info_lines, warning_lines = do_request(now)
        self.assertFalse(info_lines)
        self.assertFalse(warning_lines)

        # reinstate conf file
        with open(conf_path, 'w') as fd:
            fd.write(conf_str)

        # jump into future, send request, config reload succeeds - logged
        now += 100
        info_lines, warning_lines = do_request(now)
        self.assertEqual('Loaded new config file %s, config unchanged'
                         % conf_path, info_lines[0])
        self.assertFalse(warning_lines)

        # jump into future, send request, config reload succeeds - not logged
        now += 100
        info_lines, warning_lines = do_request(now)
        self.assertFalse(info_lines)
        self.assertFalse(warning_lines)

    def test_config_file_reload_disabled(self):
        # verify that conf file options are not periodically reloaded when
        # interval is zero
        now = time.time()
        filter_conf = {'swift_dir': self.tempdir,
                       'requests_per_device_per_second': "1.3",
                       'requests_per_device_rate_buffer': "2.4",
                       'config_reload_interval': 0}
        conf_path = os.path.join(self.tempdir, 'backend-ratelimit.conf')
        with open(conf_path, 'w') as fd:
            fd.write('[backend_ratelimit]\n'
                     'requests_per_device_per_second = 12.3\n')
        factory = backend_ratelimit.filter_factory(filter_conf)
        with mock.patch('swift.common.middleware.backend_ratelimit.time.time',
                        return_value=now):
            rl = factory(self.swift)
        exp_req_per_dev_per_sec = dict(self.default_req_per_dev_per_sec)
        exp_req_per_dev_per_sec[None] = 12.3
        self.assertEqual(exp_req_per_dev_per_sec,
                         rl.requests_per_device_per_second)
        self.assertEqual(2.4, rl.requests_per_device_rate_buffer)

        with open(conf_path, 'w') as fd:
            fd.write('[backend_ratelimit]\n'
                     'requests_per_device_per_second = 29.3\n')

        req = Request.blank('/sda1/99/a/c/o')
        self.swift.register(req.method, req.path, HTTPOk, {})
        # jump way into the future...
        with mock.patch('swift.common.middleware.backend_ratelimit.time.time',
                        return_value=now + 100000):
            resp = req.get_response(rl)
        self.assertEqual(200, resp.status_int)
        # no change
        exp_req_per_dev_per_sec = dict(self.default_req_per_dev_per_sec)
        exp_req_per_dev_per_sec[None] = 12.3
        self.assertEqual(exp_req_per_dev_per_sec,
                         rl.requests_per_device_per_second)
        self.assertEqual(2.4, rl.requests_per_device_rate_buffer)

    def _do_test_ratelimit(self, method, req_per_sec, rate_buffer,
                           extra_conf=None):
        # send 20 requests, time increments by 0.01 between each request
        start = time.time()
        fake_time = [start]

        def mock_time():
            return fake_time[0]

        app = FakeSwift()
        logger = debug_logger()
        # apply a ratelimit
        conf = {'swift_dir': self.tempdir,
                'requests_per_device_per_second': req_per_sec,
                'requests_per_device_rate_buffer': rate_buffer}
        if extra_conf:
            conf.update(extra_conf)
        rl = BackendRateLimitMiddleware(app, conf, logger)
        success = defaultdict(int)
        ratelimited = 0

        with mock.patch('swift.common.utils.time.time', mock_time):
            for i in range(20):
                for dev in ['sda1', 'sda2', 'sda3']:
                    req = Request.blank('/%s/99/a/c/o' % dev,
                                        environ={'REQUEST_METHOD': method})
                    app.register(method, req.path, HTTPOk, {})
                    resp = req.get_response(rl)
                    if resp.status_int == 200:
                        success[dev] += 1
                    else:
                        self.assertEqual(529, resp.status_int)
                        self.assertTrue(resp.status.startswith(
                            '529 Too Many Backend Requests'))
                        ratelimited += 1
                fake_time[0] += 0.01
        self.assertEqual(
            ratelimited,
            logger.statsd_client.get_increment_counts().get(
                'backend.ratelimit', 0))
        return success

    def test_method_ratelimited(self):
        def do_test_ratelimit(method):
            # no rate-limiting
            success_per_dev = self._do_test_ratelimit(method, 0, 0)
            self.assertEqual([20] * 3, list(success_per_dev.values()))

            # global rate-limited
            success_per_dev = self._do_test_ratelimit(method, 1, 0)
            self.assertEqual([1] * 3, list(success_per_dev.values()))

            success_per_dev = self._do_test_ratelimit(method, 10, 0)
            self.assertEqual([2] * 3, list(success_per_dev.values()))

            success_per_dev = self._do_test_ratelimit(method, 101, 0)
            self.assertEqual([20] * 3, list(success_per_dev.values()))

            # startup burst of 1 seconds allowance plus current allowance...
            success_per_dev = self._do_test_ratelimit(method, 1, 1)
            self.assertEqual([2] * 3, list(success_per_dev.values()))
            success_per_dev = self._do_test_ratelimit(method, 10, 1)
            self.assertEqual([12] * 3, list(success_per_dev.values()))

            # method rate-limited
            extra_conf = {
                '%s_requests_per_device_per_second' % method.lower(): 1
            }
            success_per_dev = self._do_test_ratelimit(method, 0, 0, extra_conf)
            self.assertEqual([1] * 3, list(success_per_dev.values()))

            # method not rate-limited, global rate limited
            extra_conf = {
                '%s_requests_per_device_per_second' % method.lower(): 100
            }
            success_per_dev = self._do_test_ratelimit(method, 1, 0, extra_conf)
            self.assertEqual([1] * 3, list(success_per_dev.values()))

        do_test_ratelimit('GET')
        do_test_ratelimit('HEAD')
        do_test_ratelimit('PUT')
        do_test_ratelimit('POST')
        do_test_ratelimit('DELETE')
        do_test_ratelimit('UPDATE')
        do_test_ratelimit('REPLICATE')

    def test_method_not_ratelimited(self):
        def do_test_no_ratelimit(method):
            # verify no rate-limiting
            success_per_dev = self._do_test_ratelimit(method, 1, 0)
            self.assertEqual([20] * 3, list(success_per_dev.values()))

        do_test_no_ratelimit('OPTIONS')
        do_test_no_ratelimit('SSYNC')

    def test_no_ratelimiting_configured(self):
        # verify shortcut path when no ratelimiting is configured
        with mock.patch(
                'swift.common.middleware.backend_ratelimit.'
                'BackendRateLimitMiddleware._is_allowed') as mock_is_allowed:
            success_per_dev = self._do_test_ratelimit('GET', 0, 0)
        self.assertEqual([20] * 3, list(success_per_dev.values()))
        mock_is_allowed.assert_not_called()

    def test_unhandled_request(self):
        app = FakeSwift()
        logger = debug_logger()
        conf = {'requests_per_device_per_second': 1,
                'requests_per_device_rate_buffer': 1}

        def do_test(path):
            rl = BackendRateLimitMiddleware(app, conf, logger)
            req = Request.blank(path)
            app.register('GET', req.path, HTTPOk, {})
            for i in range(10):
                resp = req.get_response(rl)
                self.assertEqual(200, resp.status_int)
            self.assertEqual(
                0, logger.statsd_client.get_increment_counts().get(
                    'backend.ratelimit', 0))

        do_test('/recon/version')
        do_test('/healthcheck')
        do_test('/v1/a/c/o')



# Copyright (c) 2011-2014 OpenStack Foundation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import json
import unittest
from datetime import datetime
import email
from unittest import mock
import time
from contextlib import contextmanager

from swift.common import swob, utils
from swift.common.http import is_success

from swift.common.middleware.s3api.s3api import filter_factory
from swift.common.middleware.s3api.etree import fromstring
from swift.common.middleware.s3api.subresource import Owner, encode_acl, \
    Grant, User, ACL, PERMISSIONS, AllUsers, AuthenticatedUsers

from test.unit.common.middleware.helpers import FakeSwift


class FakeAuthApp(object):
    container_existence_skip_cache = 0.0
    account_existence_skip_cache = 0.0

    def __init__(self, app):
        self.remote_user = 'authorized'
        self.app = app

    def _update_s3_path_info(self, env):
        """
        For S3 requests, Swift auth middleware replaces a user name in
        env['PATH_INFO'] with a valid tenant id.
        E.g. '/v1/test:tester/bucket/object' will become
        '/v1/AUTH_test/bucket/object'. This method emulates the behavior.
        """
        tenant_user = swob.str_to_wsgi(env['s3api.auth_details']['access_key'])
        tenant, user = tenant_user.rsplit(':', 1)

        path = env['PATH_INFO']
        # Make sure it's valid WSGI
        swob.wsgi_to_str(path)
        env['PATH_INFO'] = path.replace(tenant_user, 'AUTH_' + tenant)

    @staticmethod
    def authorize_cb(req):
        # Assume swift owner, if not yet set
        req.environ.setdefault('swift_owner', True)
        # But then default to blocking authz, to ensure we've replaced
        # the default auth system
        return swob.HTTPForbidden(request=req)

    def handle(self, env):
        if 's3api.auth_details' in env:
            self._update_s3_path_info(env)
        else:
            return

        if self.remote_user:
            env['REMOTE_USER'] = self.remote_user

        if env['REQUEST_METHOD'] == 'TEST':
            env['swift.authorize'] = self.authorize_cb
        else:
            env['swift.authorize'] = lambda req: None

        if 'swift.authorize_override' in env:
            return

    def __call__(self, env, start_response):
        self.handle(env)
        return self.app(env, start_response)


class S3ApiTestCase(unittest.TestCase):

    def __init__(self, name):
        unittest.TestCase.__init__(self, name)

    def _wrap_app(self, app):
        return FakeAuthApp(app)

    def setUp(self):
        # setup default config dict
        self.conf = {
            'allow_no_owner': False,
            'location': 'us-east-1',
            'dns_compliant_bucket_names': True,
            'max_bucket_listing': 1000,
            'max_parts_listing': 1000,
            'max_multi_delete_objects': 1000,
            's3_acl': False,
            'storage_domain': 'localhost',
            'auth_pipeline_check': True,
            'max_upload_part_num': 10000,
            'check_bucket_owner': False,
            'force_swift_request_proxy_log': False,
            'allow_multipart_uploads': True,
            'min_segment_size': 5242880,
            'log_level': 'debug'
        }

        # note: self.conf has no __file__ key so check_pipeline will be skipped
        # when constructing self.s3api
        self.swift = FakeSwift()
        self.app = self._wrap_app(self.swift)
        self.app._pipeline_final_app = self.swift
        self.s3api = filter_factory({}, **self.conf)(self.app)
        self.logger = self.s3api.logger = self.swift.logger

        # if you change the registered acl response for /bucket or
        # /bucket/object tearDown will complain at you; you can set this to
        # True in order to indicate you know what you're doing
        self.s3acl_response_modified = False

        self.swift.register('HEAD', '/v1/AUTH_test',
                            swob.HTTPOk, {}, None)
        self.swift.register('HEAD', '/v1/AUTH_test/bucket',
                            swob.HTTPNoContent, {}, None)
        self.swift.register('PUT', '/v1/AUTH_test/bucket',
                            swob.HTTPCreated, {}, None)
        self.swift.register('POST', '/v1/AUTH_test/bucket',
                            swob.HTTPNoContent, {}, None)
        self.swift.register('DELETE', '/v1/AUTH_test/bucket',
                            swob.HTTPNoContent, {}, None)
        self.swift.register('GET', '/v1/AUTH_test/bucket/object',
                            swob.HTTPOk, {'etag': 'object etag'}, "")
        self.swift.register('PUT', '/v1/AUTH_test/bucket/object',
                            swob.HTTPCreated, {'etag': 'object etag'}, None)
        self.swift.register('DELETE', '/v1/AUTH_test/bucket/object',
                            swob.HTTPNoContent, {}, None)

        self.mock_get_swift_info_result = {'object_versioning': {}}
        for s3api_path in (
            'controllers.obj',
            'controllers.bucket',
            'controllers.multi_delete',
            'controllers.versioning',
        ):
            patcher = mock.patch(
                'swift.common.middleware.s3api.%s.get_swift_info' % s3api_path,
                return_value=self.mock_get_swift_info_result)
            patcher.start()
            self.addCleanup(patcher.stop)

    def _register_bucket_policy_index_head(self, bucket, bucket_policy_index):
        # register bucket HEAD response with given policy index header
        headers = {'X-Backend-Storage-Policy-Index': str(bucket_policy_index)}
        self.swift.register('HEAD', '/v1/AUTH_test/' + bucket,
                            swob.HTTPNoContent, headers)

    def _assert_policy_index(self, req_headers, resp_headers, policy_index):
        self.assertNotIn('X-Backend-Storage-Policy-Index', req_headers)
        self.assertEqual(resp_headers.get('X-Backend-Storage-Policy-Index'),
                         str(policy_index))

    def _get_error_code(self, body):
        elem = fromstring(body, 'Error')
        return elem.find('./Code').text

    def _get_error_message(self, body):
        elem = fromstring(body, 'Error')
        return elem.find('./Message').text

    def _test_method_error(self, method, path, response_class, headers={},
                           env={}, expected_xml_tags=None,
                           expected_status=None):
        if not path.startswith('/'):
            path = '/' + path  # add a missing slash before the path

        uri = '/v1/AUTH_test'
        if path != '/':
            uri += path

        self.swift.register(method, uri, response_class, headers, None)
        headers.update({'Authorization': 'AWS test:tester:hmac',
                        'Date': self.get_date_header()})
        env.update({'REQUEST_METHOD': method})
        req = swob.Request.blank(path, environ=env, headers=headers)
        status, headers, body = self.call_s3api(req)
        if expected_status is not None:
            self.assertEqual(status, expected_status)
        if expected_xml_tags is not None:
            elem = fromstring(body, 'Error')
            self.assertEqual(set(expected_xml_tags),
                             {x.tag for x in elem})
        return self._get_error_code(body)

    def get_date_header(self, skew=0):
        # email.utils.formatdate returns utc timestamp in default
        return email.utils.formatdate(time.time() + skew)

    def get_v4_amz_date_header(self, offset=None):
        when = datetime.now(utils.UTC)
        if offset is not None:
            when += offset
        return when.strftime('%Y%m%dT%H%M%SZ')

    def call_app(self, req, app=None, expect_exception=False):
        if app is None:
            app = self.app

        req.headers.setdefault("User-Agent", "Mozzarella Foxfire")

        status = [None]
        headers = [None]

        def start_response(s, h, ei=None):
            status[0] = s
            headers[0] = swob.HeaderKeyDict(h)

        body_iter = app(req.environ, start_response)
        body = b''
        caught_exc = None
        try:
            for chunk in body_iter:
                body += chunk
        except Exception as exc:
            if expect_exception:
                caught_exc = exc
            else:
                raise

        if expect_exception:
            return status[0], headers[0], body, caught_exc
        else:
            return status[0], headers[0], body

    @contextmanager
    def stubbed_container_info(self, versioning_enabled=False):
        """
        some tests might want to opt-out of container_info HEAD requests; e.g.

        with self.stubbed_container_info():
            status, headers, body = self.call_s3api(req)
        """
        fake_info = {'status': 204}
        if versioning_enabled:
            fake_info['sysmeta'] = {
                'versions-container': '\x00versions\x00bucket',
            }

        with mock.patch('swift.common.middleware.s3api.s3request.'
                        'get_container_info', return_value=fake_info):
            yield

    def call_s3api(self, req, **kwargs):
        return self.call_app(req, app=self.s3api, **kwargs)


def _gen_test_headers(owner, grants=[], resource='container'):
    if not grants:
        grants = [Grant(User('test:tester'), 'FULL_CONTROL')]
    return encode_acl(resource, ACL(owner, grants))


def _gen_grant(permission):
    # generate Grant with a grantee named by "permission"
    account_name = '%s:%s' % ('test', permission.lower())
    return Grant(User(account_name), permission)


class S3ApiTestCaseAcl(S3ApiTestCase):

    def setUp(self):
        super(S3ApiTestCaseAcl, self).setUp()
        self.s3api.conf.s3_acl = True

        # some extra buckets for s3acl tests
        buckets = ['bucket', 'public', 'authenticated']
        for bucket in buckets:
            path = '/v1/AUTH_test/' + bucket
            self.swift.register('HEAD', path, swob.HTTPNoContent, {}, None),
            self.swift.register('GET', path, swob.HTTPOk, {}, json.dumps([])),

        # setup sticky ACL headers...
        self.grants = [_gen_grant(perm) for perm in PERMISSIONS]
        self.default_owner = Owner('test:tester', 'test:tester')
        container_headers = _gen_test_headers(self.default_owner, self.grants)
        object_headers = _gen_test_headers(
            self.default_owner, self.grants, 'object')
        public_headers = _gen_test_headers(
            self.default_owner, [Grant(AllUsers(), 'READ')])
        authenticated_headers = _gen_test_headers(
            self.default_owner, [Grant(AuthenticatedUsers(), 'READ')],
            'bucket')

        sticky_s3acl_headers = {
            '/v1/AUTH_test/bucket': container_headers,
            '/v1/AUTH_test/bucket+segments': container_headers,
            '/v1/AUTH_test/bucket/object': object_headers,
            '/v1/AUTH_test/public': public_headers,
            '/v1/AUTH_test/authenticated': authenticated_headers,
        }
        for path, headers in sticky_s3acl_headers.items():
            self.swift.update_sticky_response_headers(path, headers)

    def tearDown(self):
        # sanity the test didn't break the the ACLs
        swift_path_acl_resp_checks = {
            '/v1/AUTH_test/bucket': (
                'X-Container-Sysmeta-S3api-Acl', '/bucket',
                swob.HTTPNoContent),
            '/v1/AUTH_test/bucket/object': (
                'X-Object-Sysmeta-S3api-Acl', '/bucket/object', swob.HTTPOk),
        }
        check_paths = []
        for swift_path, (acl, check, resp_class) in \
                swift_path_acl_resp_checks.items():
            if self.s3acl_response_modified:
                # this is expected to reset back to the original sticky headers
                self.swift.register('HEAD', swift_path, resp_class, {}, None)
            req = swob.Request.blank(swift_path, method='HEAD')
            status, headers, body = self.call_app(req)
            if is_success(int(status.split()[0])):
                self.assertIn(acl, headers,
                              'In tearDown it seems the test (accidently?) '
                              'removed the ACL on %s' % swift_path)
                check_paths.append(check)
            else:
                self.fail('test changed resp for %s' % swift_path)
        account_expected = {
            'test:tester': 200,
            'test:other': 403,
        }
        for account, expected in account_expected.items():
            for path in check_paths:
                req = swob.Request.blank(path, method='HEAD', headers={
                    'Authorization': 'AWS %s:hmac' % account,
                    'Date': self.get_date_header()})
                status, headers, body = self.call_s3api(req)
                self.assertEqual(int(status.split()[0]), expected,
                                 'In tearDown it seems the test (accidently?) '
                                 'broke ACL access for %s to %s' % (
                                     account, path))



# Copyright (c) 2010-2012 OpenStack Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import array
import contextlib

import pickle  # nosec: B403
import json
from collections import defaultdict
from gzip import GzipFile
from os.path import getmtime
import struct
from time import time
import os
from itertools import chain, count
from tempfile import NamedTemporaryFile
import sys
import zlib

from swift.common.exceptions import RingLoadError
from swift.common.utils import hash_path, validate_configuration, md5
from swift.common.ring.utils import tiers_for_dev


DEFAULT_RELOAD_TIME = 15


def calc_replica_count(replica2part2dev_id):
    if not replica2part2dev_id:
        return 0
    base = len(replica2part2dev_id) - 1
    extra = 1.0 * len(replica2part2dev_id[-1]) / len(replica2part2dev_id[0])
    return base + extra


def normalize_devices(devs):
    # NOTE(akscram): Replication parameters like replication_ip
    #                and replication_port are required for
    #                replication process. An old replication
    #                ring doesn't contain this parameters into
    #                device. Old-style pickled rings won't have
    #                region information.
    for dev in devs:
        if dev is None:
            continue
        dev.setdefault('region', 1)
        if 'ip' in dev:
            dev.setdefault('replication_ip', dev['ip'])
        if 'port' in dev:
            dev.setdefault('replication_port', dev['port'])


class RingReader(object):
    chunk_size = 2 ** 16

    def __init__(self, filename):
        self.fp = open(filename, 'rb')
        self._reset()

    def _reset(self):
        self._buffer = b''
        self.size = 0
        self.raw_size = 0
        self._md5 = md5(usedforsecurity=False)
        self._decomp = zlib.decompressobj(32 + zlib.MAX_WBITS)

    @property
    def close(self):
        return self.fp.close

    def seek(self, pos, ref=0):
        if (pos, ref) != (0, 0):
            raise NotImplementedError
        self._reset()
        return self.fp.seek(pos, ref)

    def _buffer_chunk(self):
        chunk = self.fp.read(self.chunk_size)
        if not chunk:
            return False
        self.size += len(chunk)
        self._md5.update(chunk)
        chunk = self._decomp.decompress(chunk)
        self.raw_size += len(chunk)
        self._buffer += chunk
        return True

    def read(self, amount=-1):
        if amount < 0:
            raise IOError("don't be greedy")

        while amount > len(self._buffer):
            if not self._buffer_chunk():
                break

        result, self._buffer = self._buffer[:amount], self._buffer[amount:]
        return result

    def readline(self):
        # apparently pickle needs this?
        while b'\n' not in self._buffer:
            if not self._buffer_chunk():
                break

        line, sep, self._buffer = self._buffer.partition(b'\n')
        return line + sep

    def readinto(self, buffer):
        chunk = self.read(len(buffer))
        buffer[:len(chunk)] = chunk
        return len(chunk)

    @property
    def md5(self):
        return self._md5.hexdigest()


class RingData(object):
    """Partitioned consistent hashing ring data (used for serialization)."""

    def __init__(self, replica2part2dev_id, devs, part_shift,
                 next_part_power=None, version=None):
        normalize_devices(devs)
        self.devs = devs
        self._replica2part2dev_id = replica2part2dev_id
        self._part_shift = part_shift
        self.next_part_power = next_part_power
        self.version = version
        self.md5 = self.size = self.raw_size = None

    @property
    def replica_count(self):
        """Number of replicas (full or partial) used in the ring."""
        return calc_replica_count(self._replica2part2dev_id)

    @classmethod
    def deserialize_v1(cls, gz_file, metadata_only=False):
        """
        Deserialize a v1 ring file into a dictionary with `devs`, `part_shift`,
        and `replica2part2dev_id` keys.

        If the optional kwarg `metadata_only` is True, then the
        `replica2part2dev_id` is not loaded and that key in the returned
        dictionary just has the value `[]`.

        :param file gz_file: An opened file-like object which has already
                             consumed the 6 bytes of magic and version.
        :param bool metadata_only: If True, only load `devs` and `part_shift`
        :returns: A dict containing `devs`, `part_shift`, and
                  `replica2part2dev_id`
        """

        json_len, = struct.unpack('!I', gz_file.read(4))
        ring_dict = json.loads(gz_file.read(json_len))
        ring_dict['replica2part2dev_id'] = []

        if metadata_only:
            return ring_dict

        byteswap = (ring_dict.get('byteorder', sys.byteorder) != sys.byteorder)

        partition_count = 1 << (32 - ring_dict['part_shift'])
        for x in range(ring_dict['replica_count']):
            part2dev = array.array('H', gz_file.read(2 * partition_count))
            if byteswap:
                part2dev.byteswap()
            ring_dict['replica2part2dev_id'].append(part2dev)

        return ring_dict

    @classmethod
    def load(cls, filename, metadata_only=False):
        """
        Load ring data from a file.

        :param filename: Path to a file serialized by the save() method.
        :param bool metadata_only: If True, only load `devs` and `part_shift`.
        :returns: A RingData instance containing the loaded data.
        """
        with contextlib.closing(RingReader(filename)) as gz_file:
            # See if the file is in the new format
            magic = gz_file.read(4)
            if magic == b'R1NG':
                format_version, = struct.unpack('!H', gz_file.read(2))
                if format_version == 1:
                    ring_data = cls.deserialize_v1(
                        gz_file, metadata_only=metadata_only)
                else:
                    raise Exception('Unknown ring format version %d' %
                                    format_version)
            else:
                # Assume old-style pickled ring
                gz_file.seek(0)
                ring_data = pickle.load(gz_file)  # nosec: B301

        if hasattr(ring_data, 'devs'):
            # pickled RingData; make sure we've got region/replication info
            normalize_devices(ring_data.devs)
        else:
            ring_data = RingData(ring_data['replica2part2dev_id'],
                                 ring_data['devs'], ring_data['part_shift'],
                                 ring_data.get('next_part_power'),
                                 ring_data.get('version'))
        for attr in ('md5', 'size', 'raw_size'):
            setattr(ring_data, attr, getattr(gz_file, attr))
        return ring_data

    def serialize_v1(self, file_obj):
        # Write out new-style serialization magic and version:
        file_obj.write(struct.pack('!4sH', b'R1NG', 1))
        ring = self.to_dict()

        # Only include next_part_power if it is set in the
        # builder, otherwise just ignore it
        _text = {'devs': ring['devs'], 'part_shift': ring['part_shift'],
                 'replica_count': len(ring['replica2part2dev_id']),
                 'byteorder': sys.byteorder}

        if ring['version'] is not None:
            _text['version'] = ring['version']

        next_part_power = ring.get('next_part_power')
        if next_part_power is not None:
            _text['next_part_power'] = next_part_power

        json_text = json.dumps(_text, sort_keys=True,
                               ensure_ascii=True).encode('ascii')
        json_len = len(json_text)
        file_obj.write(struct.pack('!I', json_len))
        file_obj.write(json_text)
        for part2dev_id in ring['replica2part2dev_id']:
            part2dev_id.tofile(file_obj)

    def save(self, filename, mtime=1300507380.0):
        """
        Serialize this RingData instance to disk.

        :param filename: File into which this instance should be serialized.
        :param mtime: time used to override mtime for gzip, default or None
                      if the caller wants to include time
        """
        # Override the timestamp so that the same ring data creates
        # the same bytes on disk. This makes a checksum comparison a
        # good way to see if two rings are identical.
        tempf = NamedTemporaryFile(dir=".", prefix=filename, delete=False)
        gz_file = GzipFile(filename, mode='wb', fileobj=tempf, mtime=mtime)
        self.serialize_v1(gz_file)
        gz_file.close()
        tempf.flush()
        os.fsync(tempf.fileno())
        tempf.close()
        os.chmod(tempf.name, 0o644)
        os.rename(tempf.name, filename)

    def to_dict(self):
        return {'devs': self.devs,
                'replica2part2dev_id': self._replica2part2dev_id,
                'part_shift': self._part_shift,
                'next_part_power': self.next_part_power,
                'version': self.version}


class Ring(object):
    """
    Partitioned consistent hashing ring.

    :param serialized_path: path to serialized RingData instance
    :param reload_time: time interval in seconds to check for a ring change
    :param ring_name: ring name string (basically specified from policy)
    :param validation_hook: hook point to validate ring configuration ontime

    :raises RingLoadError: if the loaded ring data violates its constraint
    """

    def __init__(self, serialized_path, reload_time=None, ring_name=None,
                 validation_hook=lambda ring_data: None):
        # can't use the ring unless HASH_PATH_SUFFIX is set
        validate_configuration()
        if ring_name:
            self.serialized_path = os.path.join(serialized_path,
                                                ring_name + '.ring.gz')
        else:
            self.serialized_path = os.path.join(serialized_path)
        self.reload_time = (DEFAULT_RELOAD_TIME if reload_time is None
                            else reload_time)
        self._validation_hook = validation_hook
        self._reload(force=True)

    def _reload(self, force=False):
        self._rtime = time() + self.reload_time
        if force or self.has_changed():
            ring_data = RingData.load(self.serialized_path)

            try:
                self._validation_hook(ring_data)
            except RingLoadError:
                if force:
                    raise
                else:
                    # In runtime reload at working server, it's ok to use old
                    # ring data if the new ring data is invalid.
                    return

            self._mtime = getmtime(self.serialized_path)
            self._devs = ring_data.devs
            self._replica2part2dev_id = ring_data._replica2part2dev_id
            self._part_shift = ring_data._part_shift
            self._rebuild_tier_data()
            self._update_bookkeeping()
            self._next_part_power = ring_data.next_part_power
            self._version = ring_data.version
            self._md5 = ring_data.md5
            self._size = ring_data.size
            self._raw_size = ring_data.raw_size

    def _update_bookkeeping(self):
        # Do this now, when we know the data has changed, rather than
        # doing it on every call to get_more_nodes().
        #
        # Since this is to speed up the finding of handoffs, we only
        # consider devices with at least one partition assigned. This
        # way, a region, zone, or server with no partitions assigned
        # does not count toward our totals, thereby keeping the early
        # bailouts in get_more_nodes() working.
        dev_ids_with_parts = set()
        for part2dev_id in self._replica2part2dev_id:
            for dev_id in part2dev_id:
                dev_ids_with_parts.add(dev_id)
        regions = set()
        zones = set()
        ips = set()
        self._num_devs = 0
        self._num_assigned_devs = 0
        self._num_weighted_devs = 0
        for dev in self._devs:
            if dev is None:
                continue
            self._num_devs += 1
            if dev.get('weight', 0) > 0:
                self._num_weighted_devs += 1
            if dev['id'] in dev_ids_with_parts:
                regions.add(dev['region'])
                zones.add((dev['region'], dev['zone']))
                ips.add((dev['region'], dev['zone'], dev['ip']))
                self._num_assigned_devs += 1
        self._num_regions = len(regions)
        self._num_zones = len(zones)
        self._num_ips = len(ips)

    @property
    def next_part_power(self):
        if time() > self._rtime:
            self._reload()
        return self._next_part_power

    @property
    def part_power(self):
        return 32 - self._part_shift

    @property
    def version(self):
        return self._version

    @property
    def md5(self):
        return self._md5

    @property
    def size(self):
        return self._size

    @property
    def raw_size(self):
        return self._raw_size

    def _rebuild_tier_data(self):
        self.tier2devs = defaultdict(list)
        for dev in self._devs:
            if not dev:
                continue
            for tier in tiers_for_dev(dev):
                self.tier2devs[tier].append(dev)

        tiers_by_length = defaultdict(list)
        for tier in self.tier2devs:
            tiers_by_length[len(tier)].append(tier)
        self.tiers_by_length = sorted(tiers_by_length.values(),
                                      key=lambda x: len(x[0]))
        for tiers in self.tiers_by_length:
            tiers.sort()

    @property
    def replica_count(self):
        """Number of replicas (full or partial) used in the ring."""
        return calc_replica_count(self._replica2part2dev_id)

    @property
    def partition_count(self):
        """Number of partitions in the ring."""
        return len(self._replica2part2dev_id[0])

    @property
    def device_count(self):
        """Number of devices in the ring."""
        return self._num_devs

    @property
    def weighted_device_count(self):
        """Number of devices with weight in the ring."""
        return self._num_weighted_devs

    @property
    def assigned_device_count(self):
        """Number of devices with assignments in the ring."""
        return self._num_assigned_devs

    @property
    def devs(self):
        """devices in the ring"""
        if time() > self._rtime:
            self._reload()
        return self._devs

    def has_changed(self):
        """
        Check to see if the ring on disk is different than the current one in
        memory.

        :returns: True if the ring on disk has changed, False otherwise
        """
        return getmtime(self.serialized_path) != self._mtime

    def _get_part_nodes(self, part):
        part_nodes = []
        seen_ids = set()
        for r2p2d in self._replica2part2dev_id:
            if part < len(r2p2d):
                dev_id = r2p2d[part]
                if dev_id not in seen_ids:
                    part_nodes.append(self.devs[dev_id])
                    seen_ids.add(dev_id)
        return [dict(node, index=i) for i, node in enumerate(part_nodes)]

    def get_part(self, account, container=None, obj=None):
        """
        Get the partition for an account/container/object.

        :param account: account name
        :param container: container name
        :param obj: object name
        :returns: the partition number
        """
        key = hash_path(account, container, obj, raw_digest=True)
        if time() > self._rtime:
            self._reload()
        part = struct.unpack_from('>I', key)[0] >> self._part_shift
        return part

    def get_part_nodes(self, part):
        """
        Get the nodes that are responsible for the partition. If one
        node is responsible for more than one replica of the same
        partition, it will only appear in the output once.

        :param part: partition to get nodes for
        :returns: list of node dicts

        See :func:`get_nodes` for a description of the node dicts.
        """

        if time() > self._rtime:
            self._reload()
        return self._get_part_nodes(part)

    def get_nodes(self, account, container=None, obj=None):
        """
        Get the partition and nodes for an account/container/object.
        If a node is responsible for more than one replica, it will
        only appear in the output once.

        :param account: account name
        :param container: container name
        :param obj: object name
        :returns: a tuple of (partition, list of node dicts)

        Each node dict will have at least the following keys:

        ======  ===============================================================
        id      unique integer identifier amongst devices
        index   offset into the primary node list for the partition
        weight  a float of the relative weight of this device as compared to
                others; this indicates how many partitions the builder will try
                to assign to this device
        zone    integer indicating which zone the device is in; a given
                partition will not be assigned to multiple devices within the
                same zone
        ip      the ip address of the device
        port    the tcp port of the device
        device  the device's name on disk (sdb1, for example)
        meta    general use 'extra' field; for example: the online date, the
                hardware description
        ======  ===============================================================
        """
        part = self.get_part(account, container, obj)
        return part, self._get_part_nodes(part)

    def get_more_nodes(self, part):
        """
        Generator to get extra nodes for a partition for hinted handoff.

        The handoff nodes will try to be in zones other than the
        primary zones, will take into account the device weights, and
        will usually keep the same sequences of handoffs even with
        ring changes.

        :param part: partition to get handoff nodes for
        :returns: generator of node dicts

        See :func:`get_nodes` for a description of the node dicts.
        """
        if time() > self._rtime:
            self._reload()
        primary_nodes = self._get_part_nodes(part)
        used = set(d['id'] for d in primary_nodes)
        index = count()
        same_regions = set(d['region'] for d in primary_nodes)
        same_zones = set((d['region'], d['zone']) for d in primary_nodes)
        same_ips = set(
            (d['region'], d['zone'], d['ip']) for d in primary_nodes)

        parts = len(self._replica2part2dev_id[0])
        part_hash = md5(str(part).encode('ascii'),
                        usedforsecurity=False).digest()
        start = struct.unpack_from('>I', part_hash)[0] >> self._part_shift
        inc = int(parts / 65536) or 1
        # Multiple loops for execution speed; the checks and bookkeeping get
        # simpler as you go along
        hit_all_regions = len(same_regions) == self._num_regions
        for handoff_part in chain(range(start, parts, inc),
                                  range(inc - ((parts - start) % inc),
                                        start, inc)):
            if hit_all_regions:
                # At this point, there are no regions left untouched, so we
                # can stop looking.
                break
            for part2dev_id in self._replica2part2dev_id:
                if handoff_part < len(part2dev_id):
                    dev_id = part2dev_id[handoff_part]
                    dev = self._devs[dev_id]
                    region = dev['region']
                    if dev_id not in used and region not in same_regions:
                        yield dict(dev, handoff_index=next(index))
                        used.add(dev_id)
                        same_regions.add(region)
                        zone = dev['zone']
                        ip = (region, zone, dev['ip'])
                        same_zones.add((region, zone))
                        same_ips.add(ip)
                        if len(same_regions) == self._num_regions:
                            hit_all_regions = True
                            break

        hit_all_zones = len(same_zones) == self._num_zones
        for handoff_part in chain(range(start, parts, inc),
                                  range(inc - ((parts - start) % inc),
                                        start, inc)):
            if hit_all_zones:
                # Much like we stopped looking for fresh regions before, we
                # can now stop looking for fresh zones; there are no more.
                break
            for part2dev_id in self._replica2part2dev_id:
                if handoff_part < len(part2dev_id):
                    dev_id = part2dev_id[handoff_part]
                    dev = self._devs[dev_id]
                    zone = (dev['region'], dev['zone'])
                    if dev_id not in used and zone not in same_zones:
                        yield dict(dev, handoff_index=next(index))
                        used.add(dev_id)
                        same_zones.add(zone)
                        ip = zone + (dev['ip'],)
                        same_ips.add(ip)
                        if len(same_zones) == self._num_zones:
                            hit_all_zones = True
                            break

        hit_all_ips = len(same_ips) == self._num_ips
        for handoff_part in chain(range(start, parts, inc),
                                  range(inc - ((parts - start) % inc),
                                        start, inc)):
            if hit_all_ips:
                # We've exhausted the pool of unused backends, so stop
                # looking.
                break
            for part2dev_id in self._replica2part2dev_id:
                if handoff_part < len(part2dev_id):
                    dev_id = part2dev_id[handoff_part]
                    dev = self._devs[dev_id]
                    ip = (dev['region'], dev['zone'], dev['ip'])
                    if dev_id not in used and ip not in same_ips:
                        yield dict(dev, handoff_index=next(index))
                        used.add(dev_id)
                        same_ips.add(ip)
                        if len(same_ips) == self._num_ips:
                            hit_all_ips = True
                            break

        hit_all_devs = len(used) == self._num_assigned_devs
        for handoff_part in chain(range(start, parts, inc),
                                  range(inc - ((parts - start) % inc),
                                        start, inc)):
            if hit_all_devs:
                # We've used every device we have, so let's stop looking for
                # unused devices now.
                break
            for part2dev_id in self._replica2part2dev_id:
                if handoff_part < len(part2dev_id):
                    dev_id = part2dev_id[handoff_part]
                    if dev_id not in used:
                        dev = self._devs[dev_id]
                        yield dict(dev, handoff_index=next(index))
                        used.add(dev_id)
                        if len(used) == self._num_assigned_devs:
                            hit_all_devs = True
                            break



# Copyright (c) 2013 OpenStack Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os

from swift.common.constraints import valid_api_version
from swift.common.container_sync_realms import ContainerSyncRealms
from swift.common.request_helpers import append_log_info
from swift.common.swob import HTTPBadRequest, HTTPUnauthorized, wsgify
from swift.common.utils import (
    config_true_value, get_logger, streq_const_time)
from swift.proxy.controllers.base import get_container_info
from swift.common.registry import register_swift_info


class ContainerSync(object):
    """
    WSGI middleware that validates an incoming container sync request
    using the container-sync-realms.conf style of container sync.
    """

    def __init__(self, app, conf, logger=None):
        self.app = app
        self.conf = conf
        self.logger = logger or get_logger(conf, log_route='container_sync')
        self.realms_conf = ContainerSyncRealms(
            os.path.join(
                conf.get('swift_dir', '/etc/swift'),
                'container-sync-realms.conf'),
            self.logger)
        self.allow_full_urls = config_true_value(
            conf.get('allow_full_urls', 'true'))
        # configure current realm/cluster for /info
        self.realm = self.cluster = None
        current = conf.get('current', None)
        if current:
            try:
                self.realm, self.cluster = (p.upper() for p in
                                            current.strip('/').split('/'))
            except ValueError:
                self.logger.error('Invalid current //REALM/CLUSTER (%s)',
                                  current)
        self.register_info()

    def register_info(self):
        dct = {}
        for realm in self.realms_conf.realms():
            clusters = self.realms_conf.clusters(realm)
            if clusters:
                dct[realm] = {'clusters': dict((c, {}) for c in clusters)}
        if self.realm and self.cluster:
            try:
                dct[self.realm]['clusters'][self.cluster]['current'] = True
            except KeyError:
                self.logger.error('Unknown current //REALM/CLUSTER (%s)',
                                  '//%s/%s' % (self.realm, self.cluster))
        register_swift_info('container_sync', realms=dct)

    @wsgify
    def __call__(self, req):
        if req.path == '/info':
            # Ensure /info requests get the freshest results
            self.register_info()
            return self.app

        try:
            (version, acc, cont, obj) = req.split_path(3, 4, True)
            bad_path = False
        except ValueError:
            bad_path = True

        # use of bad_path bool is to avoid recursive tracebacks
        if bad_path or not valid_api_version(version):
            return self.app

        # validate container-sync metdata update
        info = get_container_info(
            req.environ, self.app, swift_source='CS')
        sync_to = req.headers.get('x-container-sync-to')
        if req.method in ('PUT', 'POST') and cont and not obj:
            versions_cont = info.get(
                'sysmeta', {}).get('versions-container')
            if sync_to and versions_cont:
                raise HTTPBadRequest(
                    'Cannot configure container sync on a container '
                    'with object versioning configured.',
                    request=req)

        if not self.allow_full_urls:
            if sync_to and not sync_to.startswith('//'):
                raise HTTPBadRequest(
                    body='Full URLs are not allowed for X-Container-Sync-To '
                         'values. Only realm values of the format '
                         '//realm/cluster/account/container are allowed.\n',
                    request=req)
        auth = req.headers.get('x-container-sync-auth')
        if auth:
            valid = False
            auth = auth.split()
            if len(auth) != 3:
                append_log_info(req.environ, 'cs:not-3-args')
            else:
                realm, nonce, sig = auth
                realm_key = self.realms_conf.key(realm)
                realm_key2 = self.realms_conf.key2(realm)
                if not realm_key:
                    append_log_info(req.environ, 'cs:no-local-realm-key')
                else:
                    user_key = info.get('sync_key')
                    if not user_key:
                        append_log_info(req.environ, 'cs:no-local-user-key')
                    else:
                        # x-timestamp headers get shunted by gatekeeper
                        if 'x-backend-inbound-x-timestamp' in req.headers:
                            req.headers['x-timestamp'] = req.headers.pop(
                                'x-backend-inbound-x-timestamp')

                        expected = self.realms_conf.get_sig(
                            req.method, req.path,
                            req.headers.get('x-timestamp', '0'), nonce,
                            realm_key, user_key)
                        expected2 = self.realms_conf.get_sig(
                            req.method, req.path,
                            req.headers.get('x-timestamp', '0'), nonce,
                            realm_key2, user_key) if realm_key2 else expected
                        if not streq_const_time(sig, expected) and \
                                not streq_const_time(sig, expected2):
                            append_log_info(req.environ, 'cs:invalid-sig')
                        else:
                            append_log_info(req.environ, 'cs:valid')
                            valid = True
            if not valid:
                exc = HTTPUnauthorized(
                    body='X-Container-Sync-Auth header not valid; '
                         'contact cluster operator for support.',
                    headers={'content-type': 'text/plain'},
                    request=req)
                exc.headers['www-authenticate'] = ' '.join([
                    'SwiftContainerSync',
                    exc.www_authenticate().split(None, 1)[1]])
                raise exc
            else:
                req.environ['swift.authorize_override'] = True
                # An SLO manifest will already be in the internal manifest
                # syntax and might be synced before its segments, so stop SLO
                # middleware from performing the usual manifest validation.
                req.environ['swift.slo_override'] = True
                # Similar arguments for static symlinks
                req.environ['swift.symlink_override'] = True

        return self.app


def filter_factory(global_conf, **local_conf):
    conf = global_conf.copy()
    conf.update(local_conf)
    register_swift_info('container_sync')

    def cache_filter(app):
        return ContainerSync(app, conf)

    return cache_filter



# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy
# of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import collections
import itertools
import json
from unittest import mock
import unittest

from swift.cli import container_deleter
from swift.common import internal_client
from swift.common import swob
from swift.common import utils

AppCall = collections.namedtuple('AppCall', [
    'method', 'path', 'query', 'headers', 'body'])


class FakeInternalClient(internal_client.InternalClient):
    def __init__(self, responses):
        self.resp_iter = iter(responses)
        self.calls = []

    def make_request(self, method, path, headers, acceptable_statuses,
                     body_file=None, params=None):
        if body_file is None:
            body = None
        else:
            body = body_file.read()
        path, _, query = path.partition('?')
        self.calls.append(AppCall(method, path, query, headers, body))
        resp = next(self.resp_iter)
        if isinstance(resp, Exception):
            raise resp
        return resp

    def __enter__(self):
        return self

    def __exit__(self, *args):
        unused_responses = [r for r in self.resp_iter]
        if unused_responses:
            raise Exception('Unused responses: %r' % unused_responses)


class TestContainerDeleter(unittest.TestCase):
    def setUp(self):
        patcher = mock.patch.object(container_deleter.time, 'time',
                                    side_effect=itertools.count())
        patcher.__enter__()
        self.addCleanup(patcher.__exit__, None, None, None)

        patcher = mock.patch.object(container_deleter, 'OBJECTS_PER_UPDATE', 5)
        patcher.__enter__()
        self.addCleanup(patcher.__exit__, None, None, None)

    def test_make_delete_jobs(self):
        ts = '1558463777.42739'
        self.assertEqual(
            container_deleter.make_delete_jobs(
                'acct', 'cont', ['obj1', 'obj2'],
                utils.Timestamp(ts)),
            [{'name': ts + '-acct/cont/obj1',
              'deleted': 0,
              'created_at': ts,
              'etag': utils.MD5_OF_EMPTY_STRING,
              'size': 0,
              'storage_policy_index': 0,
              'content_type': 'application/async-deleted'},
             {'name': ts + '-acct/cont/obj2',
              'deleted': 0,
              'created_at': ts,
              'etag': utils.MD5_OF_EMPTY_STRING,
              'size': 0,
              'storage_policy_index': 0,
              'content_type': 'application/async-deleted'}])

    def test_make_delete_jobs_native_utf8(self):
        ts = '1558463777.42739'
        uacct = acct = u'acct-\U0001f334'
        ucont = cont = u'cont-\N{SNOWMAN}'
        uobj1 = obj1 = u'obj-\N{GREEK CAPITAL LETTER ALPHA}'
        uobj2 = obj2 = u'/obj-\N{GREEK CAPITAL LETTER OMEGA}'
        self.assertEqual(
            container_deleter.make_delete_jobs(
                acct, cont, [obj1, obj2], utils.Timestamp(ts)),
            [{'name': u'%s-%s/%s/%s' % (ts, uacct, ucont, uobj1),
              'deleted': 0,
              'created_at': ts,
              'etag': utils.MD5_OF_EMPTY_STRING,
              'size': 0,
              'storage_policy_index': 0,
              'content_type': 'application/async-deleted'},
             {'name': u'%s-%s/%s/%s' % (ts, uacct, ucont, uobj2),
              'deleted': 0,
              'created_at': ts,
              'etag': utils.MD5_OF_EMPTY_STRING,
              'size': 0,
              'storage_policy_index': 0,
              'content_type': 'application/async-deleted'}])

    def test_make_delete_jobs_unicode_utf8(self):
        ts = '1558463777.42739'
        acct = u'acct-\U0001f334'
        cont = u'cont-\N{SNOWMAN}'
        obj1 = u'obj-\N{GREEK CAPITAL LETTER ALPHA}'
        obj2 = u'obj-\N{GREEK CAPITAL LETTER OMEGA}'
        self.assertEqual(
            container_deleter.make_delete_jobs(
                acct, cont, [obj1, obj2], utils.Timestamp(ts)),
            [{'name': u'%s-%s/%s/%s' % (ts, acct, cont, obj1),
              'deleted': 0,
              'created_at': ts,
              'etag': utils.MD5_OF_EMPTY_STRING,
              'size': 0,
              'storage_policy_index': 0,
              'content_type': 'application/async-deleted'},
             {'name': u'%s-%s/%s/%s' % (ts, acct, cont, obj2),
              'deleted': 0,
              'created_at': ts,
              'etag': utils.MD5_OF_EMPTY_STRING,
              'size': 0,
              'storage_policy_index': 0,
              'content_type': 'application/async-deleted'}])

    def test_mark_for_deletion_empty_no_yield(self):
        with FakeInternalClient([
            swob.Response(json.dumps([
            ])),
        ]) as swift:
            self.assertEqual(container_deleter.mark_for_deletion(
                swift,
                'account',
                'container',
                'marker',
                'end',
                'prefix',
                timestamp=None,
                yield_time=None,
            ), 0)
            self.assertEqual(swift.calls, [
                ('GET', '/v1/account/container',
                 'format=json&marker=marker&end_marker=end&prefix=prefix',
                 {}, None),
            ])

    def test_mark_for_deletion_empty_with_yield(self):
        with FakeInternalClient([
            swob.Response(json.dumps([
            ])),
        ]) as swift:
            self.assertEqual(list(container_deleter.mark_for_deletion(
                swift,
                'account',
                'container',
                'marker',
                'end',
                'prefix',
                timestamp=None,
                yield_time=0.5,
            )), [(0, None)])
            self.assertEqual(swift.calls, [
                ('GET', '/v1/account/container',
                 'format=json&marker=marker&end_marker=end&prefix=prefix',
                 {}, None),
            ])

    def test_mark_for_deletion_one_update_no_yield(self):
        ts = '1558463777.42739'
        with FakeInternalClient([
            swob.Response(json.dumps([
                {'name': '/obj1'},
                {'name': 'obj2'},
                {'name': 'obj3'},
            ])),
            swob.Response(json.dumps([
            ])),
            swob.Response(status=202),
        ]) as swift:
            self.assertEqual(container_deleter.mark_for_deletion(
                swift,
                'account',
                'container',
                '',
                '',
                '',
                timestamp=utils.Timestamp(ts),
                yield_time=None,
            ), 3)
            self.assertEqual(swift.calls, [
                ('GET', '/v1/account/container',
                 'format=json&marker=&end_marker=&prefix=', {}, None),
                ('GET', '/v1/account/container',
                 'format=json&marker=obj3&end_marker=&prefix=', {}, None),
                ('UPDATE', '/v1/.expiring_objects/' + ts.split('.')[0], '', {
                    'X-Backend-Allow-Private-Methods': 'True',
                    'X-Backend-Storage-Policy-Index': '0',
                    'X-Timestamp': ts}, mock.ANY),
            ])
            self.assertEqual(
                json.loads(swift.calls[-1].body),
                container_deleter.make_delete_jobs(
                    'account', 'container', ['/obj1', 'obj2', 'obj3'],
                    utils.Timestamp(ts)
                )
            )

    def test_mark_for_deletion_two_updates_with_yield(self):
        ts = '1558463777.42739'
        with FakeInternalClient([
            swob.Response(json.dumps([
                {'name': 'obj1'},
                {'name': 'obj2'},
                {'name': 'obj3'},
                {'name': u'obj4-\N{SNOWMAN}'},
                {'name': 'obj5'},
                {'name': 'obj6'},
            ])),
            swob.Response(status=202),
            swob.Response(json.dumps([
            ])),
            swob.Response(status=202),
        ]) as swift:
            self.assertEqual(list(container_deleter.mark_for_deletion(
                swift,
                'account',
                'container',
                '',
                'end',
                'pre',
                timestamp=utils.Timestamp(ts),
                yield_time=0,
            )), [(5, 'obj5'), (6, 'obj6'), (6, None)])
            self.assertEqual(swift.calls, [
                ('GET', '/v1/account/container',
                 'format=json&marker=&end_marker=end&prefix=pre', {}, None),
                ('UPDATE', '/v1/.expiring_objects/' + ts.split('.')[0], '', {
                    'X-Backend-Allow-Private-Methods': 'True',
                    'X-Backend-Storage-Policy-Index': '0',
                    'X-Timestamp': ts}, mock.ANY),
                ('GET', '/v1/account/container',
                 'format=json&marker=obj6&end_marker=end&prefix=pre',
                 {}, None),
                ('UPDATE', '/v1/.expiring_objects/' + ts.split('.')[0], '', {
                    'X-Backend-Allow-Private-Methods': 'True',
                    'X-Backend-Storage-Policy-Index': '0',
                    'X-Timestamp': ts}, mock.ANY),
            ])
            self.assertEqual(
                json.loads(swift.calls[-3].body),
                container_deleter.make_delete_jobs(
                    'account', 'container',
                    ['obj1', 'obj2', 'obj3', u'obj4-\N{SNOWMAN}', 'obj5'],
                    utils.Timestamp(ts)
                )
            )
            self.assertEqual(
                json.loads(swift.calls[-1].body),
                container_deleter.make_delete_jobs(
                    'account', 'container', ['obj6'],
                    utils.Timestamp(ts)
                )
            )

    def test_init_internal_client_log_name(self):
        with mock.patch(
                'swift.cli.container_deleter.InternalClient') \
                as mock_ic:
            container_deleter.main(['a', 'c', '--request-tries', '2'])
        mock_ic.assert_called_once_with(
            '/etc/swift/internal-client.conf',
            'Swift Container Deleter', 2,
            global_conf={'log_name': 'container-deleter-ic'})



# Copyright (c) 2010-2015 OpenStack Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest
import os

from swift.common.base_storage_server import BaseStorageServer
from tempfile import mkdtemp
from swift import __version__ as swift_version
from swift.common.swob import Request
from swift.common.utils import get_logger, public, replication
from shutil import rmtree


class FakeOPTIONS(BaseStorageServer):

    server_type = 'test-server'

    def __init__(self, conf, logger=None):
        super(FakeOPTIONS, self).__init__(conf)
        self.logger = logger or get_logger(conf, log_route='test-server')


class FakeANOTHER(FakeOPTIONS):

    @public
    def ANOTHER(self):
        """this is to test adding to allowed_methods"""
        pass

    @replication
    @public
    def REPLICATE(self):
        """this is to test replication_server"""
        pass

    @public
    @replication
    def REPLICATE2(self):
        """this is to test replication_server"""
        pass


class TestBaseStorageServer(unittest.TestCase):
    """Test swift.common.base_storage_server"""

    def setUp(self):
        self.tmpdir = mkdtemp()
        self.testdir = os.path.join(self.tmpdir,
                                    'tmp_test_base_storage_server')

    def tearDown(self):
        """Tear down for testing swift.common.base_storage_server"""
        rmtree(self.tmpdir)

    def test_server_type(self):
        conf = {'devices': self.testdir, 'mount_check': 'false'}
        baseserver = BaseStorageServer(conf)
        msg = 'Storage nodes have not implemented the Server type.'
        try:
            baseserver.server_type
        except NotImplementedError as e:
            self.assertEqual(str(e), msg)

    def test_allowed_methods(self):
        conf = {'devices': self.testdir, 'mount_check': 'false',
                'replication_server': 'false'}

        # test what's available in the base class
        allowed_methods_test = FakeOPTIONS(conf).allowed_methods
        self.assertEqual(allowed_methods_test, ['OPTIONS'])

        # test that a subclass can add allowed methods
        allowed_methods_test = FakeANOTHER(conf).allowed_methods
        allowed_methods_test.sort()
        self.assertEqual(allowed_methods_test, [
            'ANOTHER', 'OPTIONS'])

        conf = {'devices': self.testdir, 'mount_check': 'false',
                'replication_server': 'true'}

        # test what's available in the base class
        allowed_methods_test = FakeOPTIONS(conf).allowed_methods
        self.assertEqual(allowed_methods_test, ['OPTIONS'])

        # test that a subclass can add allowed methods
        allowed_methods_test = FakeANOTHER(conf).allowed_methods
        self.assertEqual(allowed_methods_test, [
            'ANOTHER', 'OPTIONS', 'REPLICATE', 'REPLICATE2'])

        conf = {'devices': self.testdir, 'mount_check': 'false'}

        # test what's available in the base class
        allowed_methods_test = FakeOPTIONS(conf).allowed_methods
        self.assertEqual(allowed_methods_test, ['OPTIONS'])

        # test that a subclass can add allowed methods
        allowed_methods_test = FakeANOTHER(conf).allowed_methods
        allowed_methods_test.sort()
        self.assertEqual(allowed_methods_test, [
            'ANOTHER', 'OPTIONS', 'REPLICATE', 'REPLICATE2'])

    def test_OPTIONS_error(self):
        msg = 'Storage nodes have not implemented the Server type.'
        conf = {'devices': self.testdir, 'mount_check': 'false',
                'replication_server': 'false'}

        baseserver = BaseStorageServer(conf)
        req = Request.blank('/sda1/p/a/c/o', {'REQUEST_METHOD': 'OPTIONS'})
        req.content_length = 0

        try:
            baseserver.OPTIONS(req)
        except NotImplementedError as e:
            self.assertEqual(str(e), msg)

    def test_OPTIONS(self):
        conf = {'devices': self.testdir, 'mount_check': 'false',
                'replication_server': 'false'}
        req = Request.blank('/sda1/p/a/c/o', {'REQUEST_METHOD': 'OPTIONS'})
        req.content_length = 0
        resp = FakeOPTIONS(conf).OPTIONS(req)
        self.assertEqual(resp.headers['Allow'], 'OPTIONS')
        self.assertEqual(resp.headers['Server'],
                         'test-server/' + swift_version)



# Copyright (c) 2010-2012 OpenStack Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest
from unittest import mock
import os

from swift.common.utils import normalize_timestamp
from swift.container import auditor
from test.debug_logger import debug_logger
from test.unit import with_tempdir
from test.unit.container import test_backend


class TestAuditorMigrations(unittest.TestCase):

    @with_tempdir
    @mock.patch('swift.common.db_auditor.dump_recon_cache')
    def test_db_migration(self, tempdir, mock_recon):
        db_path = os.path.join(tempdir, 'sda', 'containers', '0', '0', '0',
                               'test.db')
        with test_backend.TestContainerBrokerBeforeSPI.old_broker() as \
                old_ContainerBroker:
            broker = old_ContainerBroker(db_path, account='a', container='c')
            broker.initialize(normalize_timestamp(0), -1)

        with broker.get() as conn:
            try:
                conn.execute('SELECT storage_policy_index '
                             'FROM container_stat')
            except Exception as err:
                self.assertTrue('no such column: storage_policy_index' in
                                str(err))
            else:
                self.fail('TestContainerBrokerBeforeSPI broker class '
                          'was already migrated')

        conf = {'devices': tempdir, 'mount_check': False}
        test_auditor = auditor.ContainerAuditor(conf, logger=debug_logger())
        test_auditor.run_once()

        broker = auditor.ContainerBroker(db_path, account='a', container='c')
        info = broker.get_info()
        expected = {
            'account': 'a',
            'container': 'c',
            'object_count': 0,
            'bytes_used': 0,
            'storage_policy_index': 0,
        }
        for k, v in expected.items():
            self.assertEqual(info[k], v)


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python

# Copyright (c) 2020 SwiftStack, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse
import json
import os
import os.path
import sys
import threading
import time
import traceback

import urllib.parse
import socketserver
import http.server

try:
    import selenium.webdriver
except ImportError:
    selenium = None
import swiftclient.client

DEFAULT_ENV = {
    'OS_AUTH_URL': os.environ.get('ST_AUTH',
                                  'http://localhost:8080/auth/v1.0'),
    'OS_USERNAME': os.environ.get('ST_USER', 'test:tester'),
    'OS_PASSWORD': os.environ.get('ST_KEY', 'testing'),
    'OS_STORAGE_URL': None,
    'S3_ENDPOINT': None,
    'S3_USER': 'test:tester',
    'S3_KEY': 'testing',
}
ENV = {key: os.environ.get(key, default)
       for key, default in DEFAULT_ENV.items()}

TEST_TIMEOUT = 120.0  # seconds
STEPS = 500


class CORSSiteHandler(http.server.SimpleHTTPRequestHandler):
    def __init__(self, *args, **kwargs):
        super().__init__(
            *args,
            directory=os.path.realpath(os.path.dirname(__file__)),
            **kwargs,
        )

    def log_message(self, fmt, *args):
        pass  # quiet, you!


class CORSSiteServer(socketserver.TCPServer):
    allow_reuse_address = True


class CORSSite(threading.Thread):
    def __init__(self, bind_port=8000):
        super().__init__()
        self.server = None
        self.bind_port = bind_port

    def run(self):
        self.server = CORSSiteServer(
            ('0.0.0.0', self.bind_port),
            CORSSiteHandler)
        self.server.serve_forever()

    def terminate(self):
        if self.server is not None:
            self.server.shutdown()
            self.join()


class Zeroes(object):
    BUF = b'\x00' * 64 * 1024

    def __init__(self, size=0):
        self.pos = 0
        self.size = size

    def __iter__(self):
        while self.pos < self.size:
            chunk = self.BUF[:self.size - self.pos]
            self.pos += len(chunk)
            yield chunk

    def __len__(self):
        return self.size


def setup(args):
    conn = swiftclient.client.Connection(
        ENV['OS_AUTH_URL'],
        ENV['OS_USERNAME'],
        ENV['OS_PASSWORD'],
        timeout=30)  # We've seen request times as high as 7-8s in the gate
    cluster_info = conn.get_capabilities()
    conn.put_container('private', {
        'X-Container-Read': '',
        'X-Container-Meta-Access-Control-Allow-Origin': '',
    })
    conn.put_container('referrer-allowed', {
        'X-Container-Read': '.r:%s' % args.hostname,
        'X-Container-Meta-Access-Control-Allow-Origin': (
            'http://%s:%d' % (args.hostname, args.port)),
    })
    conn.put_container('other-referrer-allowed', {
        'X-Container-Read': '.r:other-host',
        'X-Container-Meta-Access-Control-Allow-Origin': 'http://other-host',
    })
    conn.put_container('public-with-cors', {
        'X-Container-Read': '.r:*,.rlistings',
        'X-Container-Meta-Access-Control-Allow-Origin': '*',
    })
    conn.put_container('private-with-cors', {
        'X-Container-Read': '',
        'X-Container-Meta-Access-Control-Allow-Origin': '*',
    })
    conn.put_container('public-no-cors', {
        'X-Container-Read': '.r:*,.rlistings',
        'X-Container-Meta-Access-Control-Allow-Origin': '',
    })
    conn.put_container('public-segments', {
        'X-Container-Read': '.r:*',
        'X-Container-Meta-Access-Control-Allow-Origin': '',
    })

    for container in ('private', 'referrer-allowed', 'other-referrer-allowed',
                      'public-with-cors', 'private-with-cors',
                      'public-no-cors'):
        conn.put_object(container, 'obj', Zeroes(1024), headers={
            'X-Object-Meta-Mtime': str(time.time())})
    for n in range(10):
        segment_etag = conn.put_object(
            'public-segments', 'seg%02d' % n, Zeroes(1024 * 1024),
            headers={'Content-Type': 'application/swiftclient-segment'})
        conn.put_object(
            'public-with-cors', 'dlo/seg%02d' % n, Zeroes(1024 * 1024),
            headers={'Content-Type': 'application/swiftclient-segment'})
    conn.put_object('public-with-cors', 'dlo-with-unlistable-segments', b'',
                    headers={'X-Object-Manifest': 'public-segments/seg'})
    conn.put_object('public-with-cors', 'dlo', b'',
                    headers={'X-Object-Manifest': 'public-with-cors/dlo/seg'})

    if 'slo' in cluster_info:
        conn.put_object('public-with-cors', 'slo', json.dumps([
            {'path': 'public-segments/seg%02d' % n, 'etag': segment_etag}
            for n in range(10)]), query_string='multipart-manifest=put')

    if 'symlink' in cluster_info:
        for tgt in ('private', 'public-with-cors', 'public-no-cors'):
            conn.put_object('public-with-cors', 'symlink-to-' + tgt, b'',
                            headers={'X-Symlink-Target': tgt + '/obj'})


def get_results_table(browser):
    result_table = browser.find_element_by_id('results')
    for row in result_table.find_elements_by_xpath('./tr'):
        cells = row.find_elements_by_xpath('td')
        yield (
            cells[0].text,
            browser.name + ': ' + cells[1].text,
            cells[2].text)


def run(args, url):
    results = []
    browsers = list(ALL_BROWSERS) if 'all' in args.browsers else args.browsers
    ran_one = False
    for browser_name in browsers:
        kwargs = {}
        try:
            options = getattr(
                selenium.webdriver, browser_name.title() + 'Options')()
            options.headless = True
            kwargs['options'] = options
        except AttributeError:
            # not all browser types have Options class
            pass

        driver = getattr(selenium.webdriver, browser_name.title())
        try:
            browser = driver(**kwargs)
        except Exception as e:
            if not any(x in str(e) for x in (
                'needs to be in PATH',
                'SafariDriver was not found',
                'OSError: [Errno 8] Exec format error',
                'safaridriver not available for download',
            )):
                traceback.print_exc()
            results.append(('SKIP', browser_name, str(e).strip()))
            continue
        ran_one = True
        try:
            browser.get(url)

            start = time.time()
            for _ in range(STEPS):
                status = browser.find_element_by_id('status').text
                if status.startswith('Complete'):
                    results.extend(get_results_table(browser))
                    break
                time.sleep(TEST_TIMEOUT / STEPS)
            else:
                try:
                    results.extend(get_results_table(browser))
                except Exception:
                    pass  # worth a shot
                # that took a sec; give it *one last chance* to succeed
                status = browser.find_element_by_id('status').text
                if not status.startswith('Complete'):
                    results.append((
                        'ERROR', browser_name, 'Timed out (%s)' % status))
                    continue
            sys.stderr.write('Tested %s in %.1fs\n' % (
                browser_name, time.time() - start))
        except Exception as e:
            results.append(('ERROR', browser_name, str(e).strip()))
        finally:
            browser.close()

    if args.output is not None:
        fp = open(args.output, 'w')
    else:
        fp = sys.stdout

    fp.write('1..%d\n' % len(results))
    rc = 0
    if not ran_one:
        rc += 1  # make sure "no tests ran" translates to "failed"
    for test, (status, name, details) in enumerate(results, start=1):
        if status == 'PASS':
            fp.write('ok %d - %s\n' % (test, name))
        elif status == 'SKIP':
            fp.write('ok %d - %s # skip %s\n' % (test, name, details))
        else:
            fp.write('not ok %d - %s\n' % (test, name))
            fp.write('  %s%s\n' % (status, ':' if details else ''))
            if details:
                fp.write(''.join(
                    '  ' + line + '\n'
                    for line in details.split('\n')))
            rc += 1

    if fp is not sys.stdout:
        fp.close()

    return rc


ALL_BROWSERS = [
    'firefox',
    'chrome',
    'safari',
    'edge',
    'ie',
]

if __name__ == '__main__':
    parser = argparse.ArgumentParser(
        formatter_class=argparse.RawDescriptionHelpFormatter,
        description='Set up and run CORS functional tests',
        epilog='''The tests consist of three parts:

setup - Create several test containers with well-known names, set appropriate
        ACLs and CORS metadata, and upload some test objects.
serve - Serve a static website on localhost which, on load, will make several
        CORS requests and verify expected behavior.
run   - Use Selenium to load the website, wait for and scrape the results,
        and output them in TAP format.

By default, perform all three parts. You can skip some or all of the parts
with the --no-setup, --no-serve, and --no-run options.
''')
    parser.add_argument('-P', '--port', type=int, default=8000)
    parser.add_argument('-H', '--hostname', default='localhost')
    parser.add_argument('--no-setup', action='store_true')
    parser.add_argument('--no-serve', action='store_true')
    parser.add_argument('--no-run', action='store_true')
    parser.add_argument('-o', '--output')
    parser.add_argument('browsers', nargs='*',
                        default='all',
                        choices=['all'] + ALL_BROWSERS)
    args = parser.parse_args()
    if not args.no_setup:
        setup(args)

    if args.no_serve:
        site = None
    else:
        site = CORSSite(args.port)

    should_run = not args.no_run
    if should_run and not selenium:
        print('Selenium not available; cannot run tests automatically')
        should_run = False

    if ENV['OS_STORAGE_URL'] is None:
        ENV['OS_STORAGE_URL'] = swiftclient.client.get_auth(
            ENV['OS_AUTH_URL'],
            ENV['OS_USERNAME'],
            ENV['OS_PASSWORD'],
            timeout=1)[0]
    if ENV['S3_ENDPOINT'] is None:
        ENV['S3_ENDPOINT'] = ENV['OS_STORAGE_URL'].partition('/v1')[0]

    url = 'http://%s:%d/#%s' % (args.hostname, args.port, '&'.join(
        '%s=%s' % (urllib.parse.quote(key), urllib.parse.quote(val))
        for key, val in ENV.items()))

    rc = 0
    if should_run:
        if site:
            site.start()
        try:
            rc = run(args, url)
        finally:
            if site:
                site.terminate()
    else:
        if site:
            print('Serving test at %s' % url)
            try:
                site.run()
            except KeyboardInterrupt:
                pass
    exit(rc)
